"""
The thin runner is used to manage the salt thin systems.

Salt Thin is a transport-less version of Salt that can be used to run routines
in a standalone way. This runner has tools which generate the standalone salt
system for easy consumption.
"""

import salt.utils.thin


def generate(
    extra_mods="",
    overwrite=False,
    so_mods="",
    absonly=True,
    compress="gzip",
):
    """
    Generate the salt-thin tarball and print the location of the tarball
    Optional additional mods to include (e.g. mako) can be supplied as a comma
    delimited string.  Permits forcing an overwrite of the output file as well.

    CLI Example:

    .. code-block:: bash

        salt-run thin.generate
        salt-run thin.generate mako
        salt-run thin.generate mako,wempy 1
        salt-run thin.generate overwrite=1
    """
    conf_mods = __opts__.get("thin_extra_mods")
    if conf_mods:
        extra_mods = ",".join([conf_mods, extra_mods])

    return salt.utils.thin.gen_thin(
        __opts__["cachedir"],
        extra_mods,
        overwrite,
        so_mods,
        absonly,
        compress,
    )


def generate_min(
    extra_mods="",
    overwrite=False,
    so_mods="",
):
    """
    Generate the salt-thin tarball and print the location of the tarball
    Optional additional mods to include (e.g. mako) can be supplied as a comma
    delimited string.  Permits forcing an overwrite of the output file as well.

    CLI Example:

    .. code-block:: bash

        salt-run thin.generate_min
    """
    conf_mods = __opts__.get("min_extra_mods")
    if conf_mods:
        extra_mods = ",".join([conf_mods, extra_mods])

    return salt.utils.thin.gen_min(
        __opts__["cachedir"],
        extra_mods,
        overwrite,
        so_mods,
    )



"""
A module for testing the logic of states and highstates on salt minions

:codeauthor:    William Cannon <william.cannon@gmail.com>
:maturity:      new

Saltcheck provides unittest like functionality requiring only the knowledge of
salt module execution and yaml. Saltcheck uses salt modules to return data, then
runs an assertion against that return. This allows for testing with all the
features included in salt modules.

In order to run state and highstate saltcheck tests, a sub-folder in the state directory
must be created and named ``saltcheck-tests``. Tests for a state should be created in files
ending in ``*.tst`` and placed in the ``saltcheck-tests`` folder. ``tst`` files are run
through the salt rendering system, enabling tests to be written in yaml (or renderer of choice),
and include jinja, as well as the usual grain and pillar information. Like states, multiple tests can
be specified in a ``tst`` file. Multiple ``tst`` files can be created in the ``saltcheck-tests``
folder, and should be named the same as the associated state. The ``id`` of a test works in the
same manner as in salt state files and should be unique and descriptive.


.. versionadded:: 3000
    The ``saltcheck-tests`` folder can be customized using the ``saltcheck_test_location`` minion
    configuration setting.  This setting is a relative path from the formula's ``salt://`` path
    to the test files.

Usage
=====

Example Default file system layout:

.. code-block:: text

    /srv/salt/apache/
        init.sls
        config.sls
        saltcheck-tests/
            init.tst
            config.tst
            deployment_validation.tst

Alternative example file system layout with custom saltcheck_test_location:

Minion configuration:
---------------------

.. code-block:: yaml

    saltcheck_test_location: tests/integration/saltcheck

Filesystem layout:
------------------

.. code-block:: text

    /srv/salt/apache/
        init.sls
        config.sls
        tests/integration/saltcheck/
            init.tst
            config.tst
            deployment_validation.tst

Tests can be run for each state by name, for all ``apache/saltcheck/*.tst``
files, or for all states assigned to the minion in top.sls. Tests may also be
created with no associated state. These tests will be run through the use of
``saltcheck.run_state_tests``, but will not be automatically run by
``saltcheck.run_highstate_tests``.

.. code-block:: bash

    salt '*' saltcheck.run_state_tests apache,apache.config
    salt '*' saltcheck.run_state_tests apache check_all=True
    salt '*' saltcheck.run_highstate_tests
    salt '*' saltcheck.run_state_tests apache.deployment_validation

Saltcheck Keywords
==================

**module_and_function:**
    (str) This is the salt module which will be run locally,
    the same as ``salt-call --local <module>``. The ``saltcheck.state_apply`` module name is
    special as it bypasses the local option in order to resolve state names when run in
    a master/minion environment.
**args:**
    (list) Optional arguments passed to the salt module
**kwargs:**
    (dict) Optional keyword arguments to be passed to the salt module
**assertion:**
    (str) One of the supported assertions and required except for ``saltcheck.state_apply``
    Tests which fail the assertion and expected_return, cause saltcheck to exit which a non-zero exit code.
**expected_return:**
    (str) Required except by ``assertEmpty``, ``assertNotEmpty``, ``assertTrue``,
    ``assertFalse``. The return of module_and_function is compared to this value in the assertion.
**assertion_section:**
    (str) Optional keyword used to parse the module_and_function return. If a salt module
    returns a dictionary as a result, the ``assertion_section`` value is used to lookup a specific value
    in that return for the assertion comparison.
**assertion_section_delimiter:**
    (str) Optional delimiter to use when splitting a nested structure.
    Defaults to ':'
**print_result:**
    (bool) Optional keyword to show results in the ``assertEqual``, ``assertNotEqual``,
    ``assertIn``, and ``assertNotIn`` output. Defaults to True.
**output_details:**
    (bool) Optional keyword to display ``module_and_function``, ``args``, ``assertion_section``,
    and assertion results text in the output. If print_result is False, assertion results will be hidden.
    This is a per test setting, but can be set globally for all tests by adding ``saltcheck_output_details: True``
    in the minion configuration file.
    Defaults to False
**pillar_data:**
    (dict) Optional keyword for passing in pillar data. Intended for use in potential test
    setup or teardown with the ``saltcheck.state_apply`` function.
**skip:**
    (bool) Optional keyword to skip running the individual test

.. versionadded:: 3000
    Multiple assertions can be run against the output of a single ``module_and_function`` call. The ``assertion``,
    ``expected_return``, ``assertion_section``, and ``assertion_section_delimiter`` keys can be placed in a list under an
    ``assertions`` key. See the multiple assertions example below.

Sample Cases/Examples
=====================

Basic Example
-------------

.. code-block:: yaml

    echo_test_hello:
      module_and_function: test.echo
      args:
        - "hello"
      kwargs:
      assertion: assertEqual
      expected_return:  'hello'

Example with jinja
------------------

.. code-block:: jinja

    {% for package in ["apache2", "openssh"] %}
    {# or another example #}
    {# for package in salt['pillar.get']("packages") #}
    test_{{ package }}_latest:
      module_and_function: pkg.upgrade_available
      args:
        - {{ package }}
      assertion: assertFalse
    {% endfor %}

Example with setup state including pillar
-----------------------------------------

.. code-block:: yaml

    setup_test_environment:
      module_and_function: saltcheck.state_apply
      args:
        - common
      pillar_data:
        data: value

    verify_vim:
      module_and_function: pkg.version
      args:
        - vim
      assertion: assertNotEmpty

Example with jinja
------------------

.. code-block:: jinja

    {% for package in ["apache2", "openssh"] %}
    {# or another example #}
    {# for package in salt['pillar.get']("packages") #}
    test_{{ package }}_latest:
      module_and_function: pkg.upgrade_available
      args:
        - {{ package }}
      assertion: assertFalse
    {% endfor %}

Example with setup state including pillar
-----------------------------------------

.. code-block:: yaml

    setup_test_environment:
      module_and_function: saltcheck.state_apply
      args:
        - common
      pillar-data:
        data: value

    verify_vim:
      module_and_function: pkg.version
      args:
        - vim
      assertion: assertNotEmpty

Example with skip
-----------------

.. code-block:: yaml

    package_latest:
      module_and_function: pkg.upgrade_available
      args:
        - apache2
      assertion: assertFalse
      skip: True

Example with assertion_section
------------------------------

.. code-block:: yaml

    validate_shell:
      module_and_function: user.info
      args:
        - root
      assertion: assertEqual
      expected_return: /bin/bash
      assertion_section: shell

Example with a nested assertion_section
---------------------------------------

.. code-block:: yaml

    validate_smb_signing:
      module_and_function: lgpo.get
      args:
        - 'Machine'
      kwargs:
        return_full_policy_names: True
      assertion: assertEqual
      expected_return: Enabled
      assertion_section: 'Computer Configuration|Microsoft network client: Digitally sign communications (always)'
      assertion_section_delimiter: '|'

Example suppressing print results
---------------------------------

.. code-block:: yaml

    validate_env_nameNode:
      module_and_function: hadoop.dfs
      args:
        - text
        - /oozie/common/env.properties
      expected_return: nameNode = hdfs://nameservice2
      assertion: assertNotIn
      print_result: False

Example with multiple assertions and output_details
---------------------------------------------------

.. code-block:: yaml

    multiple_validations:
      module_and_function: network.netstat
      assertions:
        - assertion: assertEqual
          assertion_section: "0:program"
          expected_return: "systemd-resolve"
        - assertion: assertEqual
          assertion_section: "0:proto"
          expected_return: "udp"
      output_details: True

Supported assertions
====================

* assertEqual
* assertNotEqual
* assertTrue
* assertFalse
* assertIn
* assertNotIn
* assertGreater
* assertGreaterEqual
* assertLess
* assertLessEqual
* assertEmpty
* assertNotEmpty

.. warning::

  The saltcheck.state_apply function is an alias for
  :py:func:`state.apply <salt.modules.state.apply>`. If using the
  :ref:`ACL system <acl-eauth>` ``saltcheck.*`` might provide more capability
  than intended if only ``saltcheck.run_state_tests`` and
  ``saltcheck.run_highstate_tests`` are needed.
"""

import copy
import logging
import multiprocessing
import os
import time

import salt.client
import salt.exceptions
import salt.utils.data
import salt.utils.files
import salt.utils.functools
import salt.utils.path
import salt.utils.platform
import salt.utils.yaml
from salt.defaults import DEFAULT_TARGET_DELIM
from salt.utils.decorators import memoize
from salt.utils.json import dumps, loads
from salt.utils.odict import OrderedDict

try:
    from junit_xml import TestCase, TestSuite

    HAS_JUNIT = True
except ImportError:
    HAS_JUNIT = False


log = logging.getLogger(__name__)

try:
    __context__
except NameError:
    __context__ = {}
__context__["global_scheck"] = None

__virtualname__ = "saltcheck"


def __virtual__():
    """
    Set the virtual pkg module if not running as a proxy
    """
    if not salt.utils.platform.is_proxy():
        return __virtualname__
    return (
        False,
        "The saltcheck execution module failed to load: only available on minions.",
    )


def run_test(**kwargs):
    """
    Execute one saltcheck test and return result

    :param keyword arg test:

    CLI Example:

    .. code-block:: bash

        salt '*' saltcheck.run_test
            test='{"module_and_function": "test.echo",
                   "assertion": "assertEqual",
                   "expected_return": "This works!",
                   "args":["This works!"] }'
    """
    # salt converts the string to a dictionary auto-magically
    scheck = SaltCheck()
    test = kwargs.get("test", None)
    if test and isinstance(test, dict):
        return scheck.run_test(test)
    else:
        return "Test argument must be a dictionary"


def state_apply(state_name, **kwargs):
    """
    Runs :py:func:`state.apply <salt.modules.state.apply>` with given options to set up test data.
    Intended to be used for optional test setup or teardown

    Reference the :py:func:`state.apply <salt.modules.state.apply>` module documentation for arguments and usage options

    CLI Example:

    .. code-block:: bash

        salt '*' saltcheck.state_apply postfix
    """
    # A new salt client is instantiated with the default configuration because the main module's
    #   client is hardcoded to local
    # minion is running with a master, a potentially non-local client is needed to lookup states
    conf_file = copy.deepcopy(__opts__["conf_file"])
    local_opts = salt.config.minion_config(conf_file)
    if "running_data/var/run/salt-minion.pid" in __opts__.get("pidfile", False):
        # Force salt-ssh minions to use local
        local_opts["file_client"] = "local"
        log.debug("Detected salt-ssh, running as local")
    caller = salt.client.Caller(mopts=local_opts)
    if kwargs:
        return caller.cmd("state.apply", state_name, **kwargs)
    else:
        return caller.cmd("state.apply", state_name)


def report_highstate_tests(saltenv=None):
    """
    Report on tests for states assigned to the minion through highstate.
    Quits with the exit code for the number of missing tests.

    CLI Example:

    .. code-block:: bash

        salt '*' saltcheck.report_highstate_tests

    .. versionadded:: 3000
    """
    if not saltenv:
        if "saltenv" in __opts__ and __opts__["saltenv"]:
            saltenv = __opts__["saltenv"]
        else:
            saltenv = "base"

    sls_list = []
    sls_list = _get_top_states(saltenv)
    stl = StateTestLoader(saltenv)
    missing_tests = 0
    states_missing_tests = []
    for state_name in sls_list:
        stl.add_test_files_for_sls(state_name, False)
        if state_name not in stl.found_states:
            missing_tests = missing_tests + 1
            states_missing_tests.append(state_name)
    __context__["retcode"] = missing_tests
    return {
        "TEST REPORT RESULTS": {
            "Missing Tests": missing_tests,
            "States missing tests": states_missing_tests,
            "States with tests": stl.found_states,
        }
    }


def run_state_tests(
    state, saltenv=None, check_all=False, only_fails=False, junit=False
):
    """
    Execute tests for a salt state and return results
    Nested states will also be tested

    :param str state: state name for which to run associated .tst test files
    :param str saltenv: optional saltenv. Defaults to base
    :param bool check_all: boolean to run all tests in state/saltcheck-tests directory
    :param bool only_fails: boolean to only print failure results
    :param bool junit: boolean to print results in junit format
        .. versionadded:: 3007.0

    CLI Example:

    .. code-block:: bash

        salt '*' saltcheck.run_state_tests postfix,common

    Tests will be run in parallel by adding "saltcheck_parallel: True" in minion config.
    When enabled, saltcheck will use up to the number of cores detected. This can be limited
    by setting the "saltcheck_processes" value to an integer to set the maximum number
    of parallel processes.
    """
    if not saltenv:
        if "saltenv" in __opts__ and __opts__["saltenv"]:
            saltenv = __opts__["saltenv"]
        else:
            saltenv = "base"

    # Use global scheck variable for reuse in each multiprocess
    __context__["global_scheck"] = SaltCheck(saltenv)

    parallel = __salt__["config.get"]("saltcheck_parallel")
    num_proc = __salt__["config.get"]("saltcheck_processes")

    stl = StateTestLoader(saltenv)
    results = OrderedDict()
    sls_list = salt.utils.args.split_input(state)
    for state_name in sls_list:
        stl.add_test_files_for_sls(state_name, check_all)
        stl.load_test_suite()
        results_dict = OrderedDict()
        # Check for situations to disable parallization
        if parallel:
            if isinstance(num_proc, float):
                num_proc = int(num_proc)

            if multiprocessing.cpu_count() < 2:
                parallel = False
                log.debug("Only 1 CPU. Disabling parallization.")
            elif num_proc == 1:
                # Don't bother with multiprocessing overhead
                parallel = False
                log.debug("Configuration limited to 1 CPU. Disabling parallization.")
            else:
                for items in stl.test_dict.values():
                    if "state.apply" in items.get("module_and_function", []):
                        # Multiprocessing doesn't ensure ordering, which state.apply
                        # might require
                        parallel = False
                        log.warning(
                            "Tests include state.apply. Disabling parallization."
                        )

        if parallel:
            if num_proc:
                pool_size = num_proc
            else:
                pool_size = min(len(stl.test_dict), multiprocessing.cpu_count())
            log.debug("Running tests in parallel with %s processes", pool_size)
            presults = multiprocessing.Pool(pool_size).map(
                func=parallel_scheck, iterable=stl.test_dict.items()
            )
            # Remove list and form expected data structure
            for item in presults:
                for key, value in item.items():
                    results_dict[key] = value
        else:
            for key, value in stl.test_dict.items():
                result = __context__["global_scheck"].run_test(value)
                results_dict[key] = result

        # If passed a duplicate state, don't overwrite with empty res
        if not results.get(state_name):
            results[state_name] = results_dict

        if junit and HAS_JUNIT:
            return _generate_junit_out_list(results)
        else:
            return _generate_out_list(results, only_fails=only_fails)


def parallel_scheck(data):
    """triggers salt-call in parallel"""
    key = data[0]
    value = data[1]
    results = {}
    results[key] = __context__["global_scheck"].run_test(value)
    return results


run_state_tests_ssh = salt.utils.functools.alias_function(
    run_state_tests, "run_state_tests_ssh"
)


def run_highstate_tests(saltenv=None, only_fails=False, junit=False):
    """
    Execute all tests for states assigned to the minion through highstate and return results

    :param str saltenv: optional saltenv. Defaults to base
    :param bool only_fails: boolean to only print failure results
    :param bool junit: boolean to print results in junit format
        .. versionadded:: 3007.0

    CLI Example:

    .. code-block:: bash

        salt '*' saltcheck.run_highstate_tests
    """
    if not saltenv:
        if "saltenv" in __opts__ and __opts__["saltenv"]:
            saltenv = __opts__["saltenv"]
        else:
            saltenv = "base"
    sls_list = []
    sls_list = _get_top_states(saltenv)
    all_states = ",".join(sls_list)

    return run_state_tests(
        all_states, saltenv=saltenv, only_fails=only_fails, junit=junit
    )


def _eval_failure_only_print(state_name, results, only_fails):
    """
    For given results, only return failures if desired
    """
    if only_fails:
        failed_tests = {}
        for test in results[state_name]:
            if results[state_name][test]["status"].startswith("Fail"):
                if failed_tests.get(state_name):
                    failed_tests[state_name].update({test: results[state_name][test]})
                else:
                    failed_tests[state_name] = {test: results[state_name][test]}
        return failed_tests
    else:
        # Show all test results
        return {state_name: results[state_name]}


def _generate_out_list(results, only_fails=False):
    """
    generate test results output list
    """
    passed = 0
    failed = 0
    skipped = 0
    missing_tests = 0
    total_time = 0.0
    out_list = []
    for state in results:
        if not results[state].items():
            missing_tests = missing_tests + 1
        else:
            for _, val in results[state].items():
                if val["status"].startswith("Pass"):
                    passed = passed + 1
                if val["status"].startswith("Fail"):
                    failed = failed + 1
                if val["status"].startswith("Skip"):
                    skipped = skipped + 1
                total_time = total_time + float(val["duration"])
        out_list.append(_eval_failure_only_print(state, results, only_fails))
    out_list = sorted(out_list, key=lambda x: sorted(x.keys()))
    out_list.append(
        {
            "TEST RESULTS": {
                "Execution Time": round(total_time, 4),
                "Passed": passed,
                "Failed": failed,
                "Skipped": skipped,
                "Missing Tests": missing_tests,
            }
        }
    )
    # Set exist code to 1 if failed tests
    # Use-cases for exist code handling of missing or skipped?
    __context__["retcode"] = 1 if failed else 0
    return out_list


def _generate_junit_out_list(results):
    """
    generates test results output list in JUnit format
    """
    total_time = 0.0
    test_cases = []
    failed = 0
    for state in results:
        if not results[state]:
            test_cases.append(TestCase("missing_test", "", "", "Test(s) Missing"))
        else:
            for name, val in sorted(results[state].items()):
                time = float(val["duration"])
                status = val["status"]
                test_cases.append(TestCase(name, "", round(time, 4)))
                if status.startswith("Fail"):
                    failed = 1
                    test_cases[len(test_cases) - 1].add_failure_info(status)
                if status.startswith("Skip"):
                    test_cases[len(test_cases) - 1].add_skipped_info(status)
                total_time = total_time + float(val["duration"])
    test_suite = TestSuite("test_results", test_cases)
    # Set exit code to 1 if failed tests
    # Use-cases for exist code handling of missing or skipped?
    __context__["retcode"] = failed
    xml_string = TestSuite.to_xml_string([test_suite])
    return xml_string


def _render_file(file_path):
    """
    call the salt utility to render a file
    """
    # salt-call slsutil.renderer /srv/salt/jinjatest/saltcheck-tests/test1.tst
    rendered = __salt__["slsutil.renderer"](
        file_path, saltenv=__context__["global_scheck"].saltenv
    )
    log.info("rendered: %s", rendered)
    return rendered


@memoize
def _is_valid_module(module):
    """
    Return a list of all modules available on minion
    """
    modules = __salt__["sys.list_modules"]()
    return bool(module in modules)


@memoize
def _is_valid_function(module_name, function):
    """
    Determine if a function is valid for a module
    """
    try:
        functions = __salt__["sys.list_functions"](module_name)
    except salt.exceptions.SaltException:
        functions = ["unable to look up functions"]
    return f"{module_name}.{function}" in functions


def _get_top_states(saltenv="base"):
    """
    Equivalent to a salt cli: salt web state.show_top
    """
    top_states = []
    top_states = __salt__["state.show_top"](saltenv=saltenv)[saltenv]
    log.debug("saltcheck for saltenv: %s found top states: %s", saltenv, top_states)
    return top_states


class SaltCheck:
    """
    This class validates and runs the saltchecks
    """

    def __init__(self, saltenv="base"):
        self.sls_list_state = []
        self.modules = []
        self.results_dict = {}
        self.results_dict_summary = {}
        self.saltenv = saltenv
        self.assertions_list = """assertEqual assertNotEqual
                                  assertTrue assertFalse
                                  assertIn assertNotIn
                                  assertGreater
                                  assertGreaterEqual
                                  assertLess assertLessEqual
                                  assertEmpty assertNotEmpty""".split()

    def _check_assertions(self, dict):
        """Validate assertion keys"""
        is_valid = True
        assertion = dict.get("assertion", None)
        # support old expected-return and newer name normalized expected_return
        exp_ret_key = any(
            key in dict.keys() for key in ["expected_return", "expected-return"]
        )
        exp_ret_val = dict.get("expected_return", dict.get("expected-return", None))

        if assertion not in self.assertions_list:
            log.error("Saltcheck: %s is not in the assertions list", assertion)
            is_valid = False

        # Only check expected returns for assertions which require them
        if assertion not in [
            "assertEmpty",
            "assertNotEmpty",
            "assertTrue",
            "assertFalse",
        ]:
            if exp_ret_key is None:
                log.error("Saltcheck: missing expected_return")
                is_valid = False
            if exp_ret_val is None:
                log.error("Saltcheck: expected_return missing a value")
                is_valid = False

        return is_valid

    def __is_valid_test(self, test_dict):
        """
        Determine if a test contains:

        - a test name
        - a valid module and function
        - a valid assertion, or valid grouping under an assertions key
        - an expected return value - if assertion type requires it
        """
        log.info("Saltcheck: validating data: %s", test_dict)
        is_valid = True
        skip = test_dict.get("skip", False)
        m_and_f = test_dict.get("module_and_function", None)

        # Running a state does not require assertions or checks
        if m_and_f == "saltcheck.state_apply":
            return is_valid

        if test_dict.get("assertions"):
            for assertion_group in test_dict.get("assertions"):
                is_valid = self._check_assertions(assertion_group)
        else:
            is_valid = self._check_assertions(test_dict)

        if m_and_f:
            module, function = m_and_f.split(".")
            if not _is_valid_module(module):
                is_valid = False
                log.error("Saltcheck: %s is not a valid module", module)
            if not _is_valid_function(module, function):
                is_valid = False
                log.error("Saltcheck: %s is not a valid function", function)
        else:
            log.error("Saltcheck: missing module_and_function")
            is_valid = False

        return is_valid

    def _call_salt_command(self, fun, args, kwargs):
        """
        Generic call of salt Caller command
        """
        # remote functions and modules won't work with local file client
        # these aren't exhaustive lists, so add to them when a module or
        # function can't operate without the remote file client
        remote_functions = ["file.check_managed_changes"]
        remote_modules = ["cp"]
        mod = fun.split(".", maxsplit=1)[0]

        conf_file = __opts__["conf_file"]
        local_opts = salt.config.minion_config(conf_file)
        # Save orginal file_client to restore after salt.client.Caller run
        orig_file_client = local_opts["file_client"]
        mlocal_opts = copy.deepcopy(local_opts)
        if fun not in remote_functions and mod not in remote_modules:
            mlocal_opts["file_client"] = "local"
        value = False
        if args and kwargs:
            value = salt.client.Caller(mopts=mlocal_opts).cmd(fun, *args, **kwargs)
        elif args and not kwargs:
            value = salt.client.Caller(mopts=mlocal_opts).cmd(fun, *args)
        elif not args and kwargs:
            value = salt.client.Caller(mopts=mlocal_opts).cmd(fun, **kwargs)
        else:
            value = salt.client.Caller(mopts=mlocal_opts).cmd(fun)
        __opts__["file_client"] = orig_file_client

        return value

    def _run_assertions(
        self,
        mod_and_func,
        args,
        data,
        module_output,
        output_details,
        assert_print_result,
    ):
        """
        Run assertion against input
        """
        value = {}

        assertion_section = data.get("assertion_section", None)
        assertion_section_delimiter = data.get(
            "assertion_section_delimiter", DEFAULT_TARGET_DELIM
        )

        if assertion_section:
            module_output = salt.utils.data.traverse_dict_and_list(
                module_output,
                assertion_section,
                default=False,
                delimiter=assertion_section_delimiter,
            )

        if mod_and_func in ["saltcheck.state_apply"]:
            assertion = "assertNotEmpty"
        else:
            assertion = data["assertion"]
        expected_return = data.get("expected_return", data.get("expected-return", None))

        if assertion not in [
            "assertIn",
            "assertNotIn",
            "assertEmpty",
            "assertNotEmpty",
            "assertTrue",
            "assertFalse",
        ]:
            expected_return = self._cast_expected_to_returned_type(
                expected_return, module_output
            )
        if assertion == "assertEqual":
            assertion_desc = "=="
            value["status"] = self.__assert_equal(
                expected_return, module_output, assert_print_result
            )
        elif assertion == "assertNotEqual":
            assertion_desc = "!="
            value["status"] = self.__assert_not_equal(
                expected_return, module_output, assert_print_result
            )
        elif assertion == "assertTrue":
            assertion_desc = "True is"
            value["status"] = self.__assert_true(module_output)
        elif assertion == "assertFalse":
            assertion_desc = "False is"
            value["status"] = self.__assert_false(module_output)
        elif assertion == "assertIn":
            assertion_desc = "IN"
            value["status"] = self.__assert_in(
                expected_return, module_output, assert_print_result
            )
        elif assertion == "assertNotIn":
            assertion_desc = "NOT IN"
            value["status"] = self.__assert_not_in(
                expected_return, module_output, assert_print_result
            )
        elif assertion == "assertGreater":
            assertion_desc = ">"
            value["status"] = self.__assert_greater(expected_return, module_output)
        elif assertion == "assertGreaterEqual":
            assertion_desc = ">="
            value["status"] = self.__assert_greater_equal(
                expected_return, module_output
            )
        elif assertion == "assertLess":
            assertion_desc = "<"
            value["status"] = self.__assert_less(expected_return, module_output)
        elif assertion == "assertLessEqual":
            assertion_desc = "<="
            value["status"] = self.__assert_less_equal(expected_return, module_output)
        elif assertion == "assertEmpty":
            assertion_desc = "IS EMPTY"
            value["status"] = self.__assert_empty(module_output)
        elif assertion == "assertNotEmpty":
            assertion_desc = "IS NOT EMPTY"
            value["status"] = self.__assert_not_empty(module_output)
        else:
            value["status"] = "Fail - bad assertion"

        if output_details:
            if assertion_section:
                assertion_section_repr_title = " {}".format("assertion_section")
                assertion_section_repr_value = f" {assertion_section}"
            else:
                assertion_section_repr_title = ""
                assertion_section_repr_value = ""
            value[f"module.function [args]{assertion_section_repr_title}"] = (
                "{} {}{}".format(
                    mod_and_func,
                    dumps(args),
                    assertion_section_repr_value,
                )
            )
            value["saltcheck assertion"] = "{}{} {}".format(
                ("" if expected_return is None else f"{expected_return} "),
                assertion_desc,
                ("hidden" if not assert_print_result else module_output),
            )

        return value

    def run_test(self, test_dict):
        """
        Run a single saltcheck test
        """
        result = {}
        start = time.time()
        global_output_details = __salt__["config.get"](
            "saltcheck_output_details", False
        )
        output_details = test_dict.get("output_details", global_output_details)
        if self.__is_valid_test(test_dict):
            skip = test_dict.get("skip", False)
            if skip:
                return {"status": "Skip", "duration": 0.0}
            mod_and_func = test_dict["module_and_function"]

            args = test_dict.get("args", None)
            kwargs = test_dict.get("kwargs", None)
            pillar_data = test_dict.get(
                "pillar_data", test_dict.get("pillar-data", None)
            )
            if pillar_data:
                if not kwargs:
                    kwargs = {}
                kwargs["pillar"] = pillar_data
            else:
                # make sure we clean pillar from previous test
                if kwargs:
                    kwargs.pop("pillar", None)

            assert_print_result = test_dict.get("print_result", True)

            actual_return = self._call_salt_command(mod_and_func, args, kwargs)

            if test_dict.get("assertions"):
                for num, assert_group in enumerate(
                    test_dict.get("assertions"), start=1
                ):
                    result[f"assertion{num}"] = self._run_assertions(
                        mod_and_func,
                        args,
                        assert_group,
                        actual_return,
                        output_details,
                        assert_print_result,
                    )
                # Walk individual assert status results to set the top level status
                # key as needed
                for k, v in copy.deepcopy(result).items():
                    if k.startswith("assertion"):
                        for assert_k, assert_v in result[k].items():
                            if assert_k.startswith("status"):
                                if result[k][assert_k] != "Pass":
                                    result["status"] = "Fail"
                if not result.get("status"):
                    result["status"] = "Pass"
            else:
                result.update(
                    self._run_assertions(
                        mod_and_func,
                        args,
                        test_dict,
                        actual_return,
                        output_details,
                        assert_print_result,
                    )
                )

        else:
            result["status"] = "Fail - invalid test"

        end = time.time()
        result["duration"] = round(end - start, 4)
        return result

    @staticmethod
    def _cast_expected_to_returned_type(expected, returned):
        """
        Determine the type of variable returned
        Cast the expected to the type of variable returned
        """
        new_expected = expected
        if returned is not None:
            ret_type = type(returned)
            if expected == "False" and ret_type == bool:
                expected = False
            try:
                new_expected = ret_type(expected)
            except ValueError:
                log.info("Unable to cast expected into type of returned")
                log.info("returned = %s", returned)
                log.info("type of returned = %s", type(returned))
                log.info("expected = %s", expected)
                log.info("type of expected = %s", type(expected))
        return new_expected

    @staticmethod
    def __assert_equal(expected, returned, assert_print_result=True):
        """
        Test if two objects are equal
        """
        result = "Pass"

        try:
            if assert_print_result:
                assert expected == returned, "{} is not equal to {}".format(
                    expected, returned
                )
            else:
                assert expected == returned, "Result is not equal"
        except AssertionError as err:
            result = "Fail: " + str(err)
        return result

    @staticmethod
    def __assert_not_equal(expected, returned, assert_print_result=True):
        """
        Test if two objects are not equal
        """
        result = "Pass"
        try:
            if assert_print_result:
                assert expected != returned, "{} is equal to {}".format(
                    expected, returned
                )
            else:
                assert expected != returned, "Result is equal"
        except AssertionError as err:
            result = "Fail: " + str(err)
        return result

    @staticmethod
    def __assert_true(returned):
        """
        Test if an boolean is True
        """
        result = "Pass"
        try:
            assert returned is True, f"{returned} not True"
        except AssertionError as err:
            result = "Fail: " + str(err)
        return result

    @staticmethod
    def __assert_false(returned):
        """
        Test if an boolean is False
        """
        result = "Pass"
        if isinstance(returned, str):
            returned = bool(returned)
        try:
            assert returned is False, f"{returned} not False"
        except AssertionError as err:
            result = "Fail: " + str(err)
        return result

    @staticmethod
    def __assert_in(expected, returned, assert_print_result=True):
        """
        Test if a value is in the list of returned values
        """
        result = "Pass"
        try:
            if assert_print_result:
                assert expected in returned, "{} not found in {}".format(
                    expected, returned
                )
            else:
                assert expected in returned, "Result not found"
        except AssertionError as err:
            result = "Fail: " + str(err)
        return result

    @staticmethod
    def __assert_not_in(expected, returned, assert_print_result=True):
        """
        Test if a value is not in the list of returned values
        """
        result = "Pass"
        try:
            if assert_print_result:
                assert expected not in returned, "{} was found in {}".format(
                    expected, returned
                )
            else:
                assert expected not in returned, "Result was found"
        except AssertionError as err:
            result = "Fail: " + str(err)
        return result

    @staticmethod
    def __assert_greater(expected, returned):
        """
        Test if a value is greater than the returned value
        """
        result = "Pass"
        try:
            assert expected > returned, f"{returned} not False"
        except AssertionError as err:
            result = "Fail: " + str(err)
        return result

    @staticmethod
    def __assert_greater_equal(expected, returned):
        """
        Test if a value is greater than or equal to the returned value
        """
        result = "Pass"
        try:
            assert expected >= returned, f"{returned} not False"
        except AssertionError as err:
            result = "Fail: " + str(err)
        return result

    @staticmethod
    def __assert_less(expected, returned):
        """
        Test if a value is less than the returned value
        """
        result = "Pass"
        try:
            assert expected < returned, f"{returned} not False"
        except AssertionError as err:
            result = "Fail: " + str(err)
        return result

    @staticmethod
    def __assert_less_equal(expected, returned):
        """
        Test if a value is less than or equal to the returned value
        """
        result = "Pass"
        try:
            assert expected <= returned, f"{returned} not False"
        except AssertionError as err:
            result = "Fail: " + str(err)
        return result

    @staticmethod
    def __assert_empty(returned):
        """
        Test if a returned value is empty
        """
        result = "Pass"
        try:
            assert not returned, f"{returned} is not empty"
        except AssertionError as err:
            result = "Fail: " + str(err)
        return result

    @staticmethod
    def __assert_not_empty(returned):
        """
        Test if a returned value is not empty
        """
        result = "Pass"
        try:
            assert returned, "value is empty"
        except AssertionError as err:
            result = "Fail: " + str(err)
        return result


class StateTestLoader:
    """
    Class loads in test files for a state
    e.g. state_dir/saltcheck-tests/[1.tst, 2.tst, 3.tst]
    """

    def __init__(self, saltenv="base"):
        self.path_type = None
        self.test_files = set()  # list of file paths
        self.test_dict = OrderedDict()
        self.saltenv = saltenv
        self.saltcheck_test_location = __salt__["config.get"](
            "saltcheck_test_location", "saltcheck-tests"
        )
        self.found_states = []

    def load_test_suite(self):
        """
        Load tests either from one file, or a set of files
        """
        self.test_dict = OrderedDict()
        for myfile in self.test_files:
            self._load_file_salt_rendered(myfile)
        self.test_files = set()

    def _load_file_salt_rendered(self, filepath):
        """
        loads in one test file
        """
        # use the salt renderer module to interpret jinja and etc
        tests = _render_file(filepath)
        # use json as a convenient way to convert the OrderedDicts from salt renderer
        mydict = loads(dumps(tests), object_pairs_hook=OrderedDict)
        for key, value in mydict.items():
            self.test_dict[key] = value
        return

    def _copy_state_files(self, sls_path, state_name, check_all):
        """
        Copy tst files for a given path and return results of the copy.
        If check_all is enabled, also add all tests found
        """
        cache_ret = []
        if state_name not in self.found_states:
            log.debug("looking in %s to cache tests", sls_path)
            cache_ret = __salt__["cp.cache_dir"](
                sls_path, saltenv=self.saltenv, include_pat="*.tst"
            )
            if cache_ret:
                if check_all:
                    log.debug("Adding all found test files: %s", cache_ret)
                    self.test_files.update(cache_ret)
                else:
                    log.debug("Marking found_state: %s", state_name)
                    self.found_states.append(state_name)
        else:
            log.debug("Not copying already found_state: %s", self.found_states)

        return cache_ret

    def _generate_sls_path(self, state_name):
        """
        For a given state_name, return list of paths to search for .tst files

        possible formula paths are then
         path/to/formula.sls
           with tests of
             path/to/saltcheck-tests/formula.tst
         path/to/formula/init.sls
           with tests of
              path/to/formula/saltcheck-tests/init.tst
         or if a custom saltcheck_test_location is used
         path/to/forumla.sls
           with tests of
              path/saltcheck_test_location/init.tst
        """

        all_sls_paths = []

        # process /patch/to/formula/saltcheck_test_location
        test_path = "salt://{}/{}".format(
            state_name.replace(".", "/"), self.saltcheck_test_location
        )
        all_sls_paths.append(test_path)

        # process /path/to/saltcheck_test_location
        sls_split = state_name.split(".")
        sls_split.pop()
        test_path = "salt://{}/{}".format(
            "/".join(sls_split), self.saltcheck_test_location
        )
        all_sls_paths.append(test_path)

        state_name_base = state_name.split(".")[0]
        test_path = f"salt://{state_name_base}/{self.saltcheck_test_location}"
        all_sls_paths.append(test_path)

        unique_paths = set(all_sls_paths)
        # Try longer (more complicated) paths before shorter simpler ones. Ensures that
        # thing/sub/saltcheck-tests/testname will be found before thing/saltcheck-tests/testname
        return list(sorted(unique_paths, key=len, reverse=True))

    @memoize
    def _get_states(self):
        """
        Returns (cached) list of states for the minion
        """
        return __salt__["cp.list_states"](saltenv=self.saltenv)

    def add_test_files_for_sls(self, sls_name, check_all=False):
        """
        Detects states used, caches needed files, and adds to test list
        """
        salt_ssh = False
        if "running_data/var/run/salt-minion.pid" in __opts__.get("pidfile", False):
            salt_ssh = True
            log.debug("Running on salt-ssh minion. Reading file %s", sls_name)
            cp_output_file = os.path.join(
                __opts__["cachedir"], "files", self.saltenv, "cp_output.txt"
            )
            with salt.utils.files.fopen(cp_output_file, "r") as fp:
                all_states = loads(salt.utils.stringutils.to_unicode(fp.read()))
        else:
            all_states = self._get_states()

        ret = []
        cached_copied_files = []
        if salt_ssh:
            # populate cached_copied_files from sent over file rather than attempting to run cp.cache_dir later
            log.debug("Running on salt-ssh minion. Populating test file results")
            state_copy_file = os.path.join(
                __opts__["cachedir"], "files", self.saltenv, sls_name + ".copy"
            )
            try:
                with salt.utils.files.fopen(state_copy_file, "r") as fp:
                    cached_copied_files.extend(
                        loads(salt.utils.stringutils.to_unicode(fp.read()))
                    )
            except OSError:
                # likely attempting to find state.nested.copy when file was sent as just state.copy
                sls_name_list = sls_name.split(".")
                sls_root_name = ".".join(sls_name_list[:-1])
                state_copy_file = os.path.join(
                    __opts__["cachedir"], "files", self.saltenv, sls_root_name + ".copy"
                )
                with salt.utils.files.fopen(state_copy_file, "r") as fp:
                    cached_copied_files.extend(
                        loads(salt.utils.stringutils.to_unicode(fp.read()))
                    )

        if sls_name in all_states:
            if salt_ssh:
                log.debug(
                    "Running on salt-ssh minion. Reading file %s", sls_name + ".low"
                )
                state_low_file = os.path.join(
                    __opts__["cachedir"], "files", self.saltenv, sls_name + ".low"
                )
                with salt.utils.files.fopen(state_low_file, "r") as fp:
                    ret = loads(salt.utils.stringutils.to_unicode(fp.read()))
            else:
                ret = __salt__["state.show_low_sls"](
                    sls_name, saltenv=self.saltenv, test=True
                )
        else:
            # passed name isn't a state, so we'll assume it is a test definition
            ret = [{"__sls__": sls_name}]

        for low_data in ret:
            if not isinstance(low_data, dict):
                log.error(
                    "low data from show_low_sls is not formed as a dict: %s", low_data
                )
                return
            this_cache_ret = None
            if "__sls__" in low_data:
                # this low data has an SLS path in it

                state_name = low_data["__sls__"]

                for sls_path in self._generate_sls_path(state_name):
                    this_cache_ret = self._copy_state_files(
                        sls_path, state_name, check_all
                    )
                    if this_cache_ret:
                        log.debug("found tests: %s", this_cache_ret)
                        cached_copied_files.extend(this_cache_ret)

                if salt_ssh:
                    if check_all:
                        # load all tests for this state on ssh minion
                        tst_files = [
                            file_string
                            for file_string in cached_copied_files
                            if file_string.endswith(".tst")
                        ]
                        self.test_files.update(tst_files)

                if not check_all:
                    # in check_all case, tests already added
                    split_sls = low_data["__sls__"].split(".")
                    sls_path_names = {
                        os.path.join(
                            os.sep.join(split_sls),
                            os.path.normpath(self.saltcheck_test_location),
                            "init.tst",
                        ),
                        os.path.join(
                            os.sep.join(split_sls[: len(split_sls) - 1]),
                            os.path.normpath(self.saltcheck_test_location),
                            f"{split_sls[-1]}.tst",
                        ),
                        os.path.join(
                            split_sls[0],
                            os.path.normpath(self.saltcheck_test_location),
                            os.sep.join(split_sls[1:-1]),
                            f"{split_sls[-1]}.tst",
                        ),
                    }
                    # for this state, find matching test files and load them
                    cached_copied_files = list(set(cached_copied_files))
                    for this_cached_test_file in cached_copied_files:
                        if this_cached_test_file.endswith(tuple(sls_path_names)):
                            self.test_files.add(this_cached_test_file)
                            cached_copied_files.remove(this_cached_test_file)
                            log.debug("Adding .tst file: %s", this_cached_test_file)



'''
This namespace is here only to provide compatibility with 0.6.x

It will be removed in 0.8.x
'''

import sys
import warnings

# load pyroute2 entry points
import pyroute2  # noqa: F401

warnings.warn(
    'pr2modules namespace is deprecated, use pyroute2 instead',
    DeprecationWarning,
)

# alias every `pyroute2` entry, in addition to the block above
#
# Bug-Url: https://github.com/svinota/pyroute2/issues/913
#
for key, value in list(sys.modules.items()):
    if key.startswith("pyroute2."):
        sys.modules[key.replace("pyroute2", "pr2modules")] = value



import struct

from pyroute2.netlink import NLMSG_DONE, nlmsg
from pyroute2.netlink.rtnl.rtmsg import rtmsg


def default_routes(data, offset, length):
    '''
    Only for RTM_NEWROUTE.

    This parser returns:

    * rtmsg() -- only for default routes (no RTA_DST)
    * nlmsg() -- NLMSG_DONE
    * None for any other messages
    '''
    # get message header
    header = dict(
        zip(
            ('length', 'type', 'flags', 'sequence_number'),
            struct.unpack_from('IHHI', data, offset),
        )
    )
    header['error'] = None
    if header['type'] == NLMSG_DONE:
        msg = nlmsg()
        msg['header'] = header
        msg.length = msg['header']['length']
        return msg

    # skip to NLA: offset + nlmsg header + rtmsg data
    cursor = offset + 28

    # iterate NLA, if meet RTA_DST -- return None (not a default route)
    while cursor < offset + length:
        nla_length, nla_type = struct.unpack_from('HH', data, cursor)
        nla_length = (nla_length + 3) & ~3  # align, page size = 4
        cursor += nla_length
        if nla_type == 1:
            return

    # no RTA_DST, a default route -- spend time to decode using the
    # standard routine
    msg = rtmsg(data, offset=offset)
    msg.decode()
    msg['header']['error'] = None  # required
    return msg



"""
System Profiler Module

Interface with macOS's command-line System Profiler utility to get
information about package receipts and installed applications.

.. versionadded:: 2015.5.0

"""

import plistlib
import subprocess

import salt.utils.path

PROFILER_BINARY = "/usr/sbin/system_profiler"


def __virtual__():
    """
    Check to see if the system_profiler binary is available
    """
    PROFILER_BINARY = salt.utils.path.which("system_profiler")

    if PROFILER_BINARY:
        return True
    return (
        False,
        "The system_profiler execution module cannot be loaded: "
        "system_profiler unavailable.",
    )


def _call_system_profiler(datatype):
    """
    Call out to system_profiler.  Return a dictionary
    of the stuff we are interested in.
    """

    p = subprocess.Popen(
        [PROFILER_BINARY, "-detailLevel", "full", "-xml", datatype],
        stdout=subprocess.PIPE,
    )
    (sysprofresults, sysprof_stderr) = p.communicate(input=None)

    plist = plistlib.readPlistFromBytes(sysprofresults)

    try:
        apps = plist[0]["_items"]
    except (IndexError, KeyError):
        apps = []

    return apps


def receipts():
    """
    Return the results of a call to
    ``system_profiler -xml -detail full SPInstallHistoryDataType``
    as a dictionary.  Top-level keys of the dictionary
    are the names of each set of install receipts, since
    there can be multiple receipts with the same name.
    Contents of each key are a list of dictionaries.

    CLI Example:

    .. code-block:: bash

        salt '*' systemprofiler.receipts
    """

    apps = _call_system_profiler("SPInstallHistoryDataType")

    appdict = {}

    for a in apps:
        details = dict(a)
        details.pop("_name")
        if "install_date" in details:
            details["install_date"] = details["install_date"].strftime(
                "%Y-%m-%d %H:%M:%S"
            )
        if "info" in details:
            try:
                details["info"] = "{}: {}".format(
                    details["info"][0], details["info"][1].strftime("%Y-%m-%d %H:%M:%S")
                )
            except (IndexError, AttributeError):
                pass

        if a["_name"] not in appdict:
            appdict[a["_name"]] = []

        appdict[a["_name"]].append(details)

    return appdict


def applications():
    """
    Return the results of a call to
    ``system_profiler -xml -detail full SPApplicationsDataType``
    as a dictionary.  Top-level keys of the dictionary
    are the names of each set of install receipts, since
    there can be multiple receipts with the same name.
    Contents of each key are a list of dictionaries.

    Note that this can take a long time depending on how many
    applications are installed on the target Mac.

    CLI Example:

    .. code-block:: bash

        salt '*' systemprofiler.applications
    """

    apps = _call_system_profiler("SPApplicationsDataType")

    appdict = {}

    for a in apps:
        details = dict(a)
        details.pop("_name")
        if "lastModified" in details:
            details["lastModified"] = details["lastModified"].strftime(
                "%Y-%m-%d %H:%M:%S"
            )
        if "info" in details:
            try:
                details["info"] = "{}: {}".format(
                    details["info"][0], details["info"][1].strftime("%Y-%m-%d %H:%M:%S")
                )
            except (IndexError, AttributeError):
                pass

        if a["_name"] not in appdict:
            appdict[a["_name"]] = []

        appdict[a["_name"]].append(details)

    return appdict



"""
Helpful decorators for module writing
"""

import errno
import inspect
import logging
import subprocess
import sys
import time
from collections import defaultdict
from functools import wraps

import salt.utils.args
import salt.utils.data
import salt.utils.versions
from salt.exceptions import (
    CommandExecutionError,
    SaltConfigurationError,
    SaltInvocationError,
)

IS_WINDOWS = False
if getattr(sys, "getwindowsversion", False):
    IS_WINDOWS = True

log = logging.getLogger(__name__)


class Depends:
    """
    This decorator will check the module when it is loaded and check that the
    dependencies passed in are in the globals of the module. If not, it will
    cause the function to be unloaded (or replaced).
    """

    # kind -> Dependency -> list of things that depend on it
    dependency_dict = defaultdict(lambda: defaultdict(dict))

    def __init__(self, *dependencies, **kwargs):
        """
        The decorator is instantiated with a list of dependencies (string of
        global name)

        An example use of this would be:

        .. code-block:: python

            @depends('modulename')
            def test():
                return 'foo'

            OR

            @depends('modulename', fallback_function=function)
            def test():
                return 'foo'

        .. code-block:: python

        This can also be done with the retcode of a command, using the
        ``retcode`` argument:

            @depends('/opt/bin/check_cmd', retcode=0)
            def test():
                return 'foo'

        It is also possible to check for any nonzero retcode using the
        ``nonzero_retcode`` argument:

            @depends('/opt/bin/check_cmd', nonzero_retcode=True)
            def test():
                return 'foo'

        .. note::
            The command must be formatted as a string, not a list of args.
            Additionally, I/O redirection and other shell-specific syntax are
            not supported since this uses shell=False when calling
            subprocess.Popen().

        """
        log.trace(
            "Depends decorator instantiated with dep list of %s and kwargs %s",
            dependencies,
            kwargs,
        )
        self.dependencies = dependencies
        self.params = kwargs

    def __call__(self, function):
        """
        The decorator is "__call__"d with the function, we take that function
        and determine which module and function name it is to store in the
        class wide dependency_dict
        """
        try:
            # This inspect call may fail under certain conditions in the loader.
            # Possibly related to a Python bug here:
            # http://bugs.python.org/issue17735
            frame = inspect.currentframe().f_back
            # due to missing *.py files under esky we cannot use inspect.getmodule
            # module name is something like salt.loaded.int.modules.test
            _, kind, mod_name = frame.f_globals["__name__"].rsplit(".", 2)
            fun_name = function.__name__
            for dep in self.dependencies:
                self.dependency_dict[kind][dep][(mod_name, fun_name)] = (
                    frame,
                    self.params,
                )
        except Exception as exc:  # pylint: disable=broad-except
            log.exception(
                "Exception encountered when attempting to inspect frame in "
                "dependency decorator"
            )
        return function

    @staticmethod
    def run_command(dependency, mod_name, func_name):
        full_name = f"{mod_name}.{func_name}"
        log.trace("Running '%s' for '%s'", dependency, full_name)
        if IS_WINDOWS:
            args = salt.utils.args.shlex_split(dependency, posix=False)
        else:
            args = salt.utils.args.shlex_split(dependency)
        log.trace("Command after shlex_split: %s", args)
        proc = subprocess.Popen(
            args, shell=False, stdout=subprocess.PIPE, stderr=subprocess.STDOUT
        )
        output = proc.communicate()[0]
        retcode = proc.returncode
        log.trace("Output from '%s': %s", dependency, output)
        log.trace("Retcode from '%s': %d", dependency, retcode)
        return retcode

    @classmethod
    def enforce_dependencies(cls, functions, kind, tgt_mod):
        """
        This is a class global method to enforce the dependencies that you
        currently know about.
        It will modify the "functions" dict and remove/replace modules that
        are missing dependencies.
        """
        for dependency, dependent_dict in cls.dependency_dict[kind].items():
            for (mod_name, func_name), (frame, params) in dependent_dict.items():
                if mod_name != tgt_mod:
                    continue
                # Imports from local context take presedence over those from the global context.
                dep_found = frame.f_locals.get(dependency) or frame.f_globals.get(
                    dependency
                )
                # Default to version ``None`` if not found, which will be less than anything.
                dep_version = getattr(dep_found, "__version__", None)
                if "retcode" in params or "nonzero_retcode" in params:
                    try:
                        retcode = cls.run_command(dependency, mod_name, func_name)
                    except OSError as exc:
                        if exc.errno == errno.ENOENT:
                            log.trace(
                                "Failed to run command %s, %s not found",
                                dependency,
                                exc.filename,
                            )
                        else:
                            log.trace("Failed to run command '%s': %s", dependency, exc)
                        retcode = -1

                    if "retcode" in params:
                        if params["retcode"] == retcode:
                            continue

                    elif "nonzero_retcode" in params:
                        if params["nonzero_retcode"]:
                            if retcode != 0:
                                continue
                        else:
                            if retcode == 0:
                                continue

                # check if dependency is loaded
                elif dependency is True:
                    log.trace(
                        "Dependency for %s.%s exists, not unloading",
                        mod_name,
                        func_name,
                    )
                    continue

                # check if you have the dependency
                elif dep_found:
                    if "version" in params:
                        if (
                            salt.utils.versions.version_cmp(
                                dep_version, params["version"]
                            )
                            >= 0
                        ):
                            log.trace(
                                "Dependency (%s) already loaded inside %s with "
                                "version (%s), required (%s), skipping",
                                dependency,
                                mod_name,
                                dep_version,
                                params["version"],
                            )
                            continue
                    else:
                        log.trace(
                            "Dependency (%s) already loaded inside %s, skipping",
                            dependency,
                            mod_name,
                        )
                        continue

                log.trace(
                    "Unloading %s.%s because dependency (%s%s) is not met",
                    mod_name,
                    func_name,
                    dependency,
                    (
                        " version {}".format(params["version"])
                        if "version" in params
                        else ""
                    ),
                )
                # if not, unload the function
                if frame:
                    try:
                        func_name = frame.f_globals["__func_alias__"][func_name]
                    except (AttributeError, KeyError):
                        pass

                    mod_key = f"{mod_name}.{func_name}"

                    # if we don't have this module loaded, skip it!
                    if mod_key not in functions:
                        continue

                    try:
                        fallback_function = params.get("fallback_function")
                        if fallback_function is not None:
                            functions[mod_key] = fallback_function
                        else:
                            del functions[mod_key]
                    except AttributeError:
                        # we already did???
                        log.trace("%s already removed, skipping", mod_key)
                        continue


depends = Depends


def timing(function):
    """
    Decorator wrapper to log execution time, for profiling purposes
    """

    @wraps(function)
    def wrapped(*args, **kwargs):
        start_time = time.time()
        ret = function(*args, **salt.utils.args.clean_kwargs(**kwargs))
        end_time = time.time()
        if function.__module__.startswith("salt.loaded.int."):
            mod_name = function.__module__[16:]
        else:
            mod_name = function.__module__
        fstr = f"Function %s.%s took %.{sys.float_info.dig}f seconds to execute"
        log.profile(fstr, mod_name, function.__name__, end_time - start_time)
        return ret

    return wrapped


def memoize(func):
    """
    Memoize aka cache the return output of a function
    given a specific set of arguments

    .. versionedited:: 2016.3.4

    Added **kwargs support.
    """
    cache = {}

    @wraps(func)
    def _memoize(*args, **kwargs):
        str_args = []
        for arg in args:
            if not isinstance(arg, str):
                str_args.append(str(arg))
            else:
                str_args.append(arg)

        args_ = ",".join(list(str_args) + [f"{k}={kwargs[k]}" for k in sorted(kwargs)])
        if args_ not in cache:
            cache[args_] = func(*args, **kwargs)
        return cache[args_]

    return _memoize


class _DeprecationDecorator:
    """
    Base mix-in class for the deprecation decorator.
    Takes care of a common functionality, used in its derivatives.
    """

    OPT_IN = 1
    OPT_OUT = 2

    def __init__(self, globals, version):
        """
        Constructor.

        :param globals: Module globals. Important for finding out replacement functions
        :param version: Expiration version
        :return:
        """
        from salt.version import SaltStackVersion, __saltstack_version__

        self._globals = globals
        self._exp_version_name = version
        self._exp_version = SaltStackVersion.from_name(self._exp_version_name)
        self._curr_version = __saltstack_version__.info
        self._raise_later = None
        self._function = None
        self._orig_f_name = None

    def _get_args(self, kwargs):
        """
        Discard all keywords which aren't function-specific from the kwargs.

        :param kwargs:
        :return:
        """
        _args = list()
        _kwargs = salt.utils.args.clean_kwargs(**kwargs)

        return _args, _kwargs

    def _call_function(self, kwargs):
        """
        Call target function that has been decorated.

        :return:
        """
        if self._raise_later:
            raise self._raise_later  # pylint: disable=E0702

        if self._function:
            args, kwargs = self._get_args(kwargs)
            try:
                return self._function(*args, **kwargs)
            except TypeError as error:
                error = str(error).replace(
                    self._function, self._orig_f_name
                )  # Hide hidden functions
                log.error(
                    'Function "%s" was not properly called: %s',
                    self._orig_f_name,
                    error,
                )
                return self._function.__doc__
            except Exception as error:  # pylint: disable=broad-except
                log.error(
                    'Unhandled exception occurred in function "%s: %s',
                    self._function.__name__,
                    error,
                )
                raise
        else:
            raise CommandExecutionError(
                "Function is deprecated, but the successor function was not found."
            )

    def __call__(self, function):
        """
        Callable method of the decorator object when
        the decorated function is gets called.

        :param function:
        :return:
        """
        self._function = function
        self._orig_f_name = self._function.__name__


class _IsDeprecated(_DeprecationDecorator):
    """
    This decorator should be used only with the deprecated functions
    to mark them as deprecated and alter its behavior a corresponding way.
    The usage is only suitable if deprecation process is renaming
    the function from one to another. In case function name or even function
    signature stays the same, please use 'with_deprecated' decorator instead.

    It has the following functionality:

    1. Put a warning level message to the log, informing that
       the deprecated function has been in use.

    2. Raise an exception, if deprecated function is being called,
       but the lifetime of it already expired.

    3. Point to the successor of the deprecated function in the
       log messages as well during the blocking it, once expired.

    Usage of this decorator as follows. In this example no successor
    is mentioned, hence the function "foo()" will be logged with the
    warning each time is called and blocked completely, once EOF of
    it is reached:

        from salt.util.decorators import is_deprecated

        @is_deprecated(globals(), "Beryllium")
        def foo():
            pass

    In the following example a successor function is mentioned, hence
    every time the function "bar()" is called, message will suggest
    to use function "baz()" instead. Once EOF is reached of the function
    "bar()", an exception will ask to use function "baz()", in order
    to continue:

        from salt.util.decorators import is_deprecated

        @is_deprecated(globals(), "Beryllium", with_successor="baz")
        def bar():
            pass

        def baz():
            pass
    """

    def __init__(self, globals, version, with_successor=None):
        """
        Constructor of the decorator 'is_deprecated'.

        :param globals: Module globals
        :param version: Version to be deprecated
        :param with_successor: Successor function (optional)
        :return:
        """
        _DeprecationDecorator.__init__(self, globals, version)
        self._successor = with_successor

    def __call__(self, function):
        """
        Callable method of the decorator object when
        the decorated function is gets called.

        :param function:
        :return:
        """
        _DeprecationDecorator.__call__(self, function)

        @wraps(function)
        def _decorate(*args, **kwargs):
            """
            Decorator function.

            :param args:
            :param kwargs:
            :return:
            """
            if self._curr_version < self._exp_version:
                msg = [
                    'The function "{f_name}" is deprecated and will '
                    'expire in version "{version_name}".'.format(
                        f_name=self._function.__name__,
                        version_name=self._exp_version_name,
                    )
                ]
                if self._successor:
                    msg.append(
                        'Use successor "{successor}" instead.'.format(
                            successor=self._successor
                        )
                    )
                log.warning(" ".join(msg))
            else:
                msg = [
                    'The lifetime of the function "{f_name}" expired.'.format(
                        f_name=self._function.__name__
                    )
                ]
                if self._successor:
                    msg.append(
                        'Please use its successor "{successor}" instead.'.format(
                            successor=self._successor
                        )
                    )
                log.warning(" ".join(msg))
                raise CommandExecutionError(" ".join(msg))
            return self._call_function(kwargs)

        return _decorate


is_deprecated = _IsDeprecated


class _WithDeprecated(_DeprecationDecorator):
    """
    This decorator should be used with the successor functions
    to mark them as a new and alter its behavior in a corresponding way.
    It is used alone if a function content or function signature
    needs to be replaced, leaving the name of the function same.
    In case function needs to be renamed or just dropped, it has
    to be used in pair with 'is_deprecated' decorator.

    It has the following functionality:

    1. Put a warning level message to the log, in case a component
       is using its deprecated version.

    2. Switch between old and new function in case an older version
       is configured for the desired use.

    3. Raise an exception, if deprecated version reached EOL and
       point out for the new version.

    Usage of this decorator as follows. If 'with_name' is not specified,
    then the name of the deprecated function is assumed with the "_" prefix.
    In this case, in order to deprecate a function, it is required:

    - Add a prefix "_" to an existing function. E.g.: "foo()" to "_foo()".

    - Implement a new function with exactly the same name, just without
      the prefix "_".

    Example:

        from salt.util.decorators import with_deprecated

        @with_deprecated(globals(), "Beryllium")
        def foo():
            "This is a new function"

        def _foo():
            "This is a deprecated function"


    In case there is a need to deprecate a function and rename it,
    the decorator should be used with the 'with_name' parameter. This
    parameter is pointing to the existing deprecated function. In this
    case deprecation process as follows:

    - Leave a deprecated function without changes, as is.

    - Implement a new function and decorate it with this decorator.

    - Set a parameter 'with_name' to the deprecated function.

    - If a new function has a different name than a deprecated,
      decorate a deprecated function with the  'is_deprecated' decorator
      in order to let the function have a deprecated behavior.

    Example:

        from salt.util.decorators import with_deprecated

        @with_deprecated(globals(), "Beryllium", with_name="an_old_function")
        def a_new_function():
            "This is a new function"

        @is_deprecated(globals(), "Beryllium", with_successor="a_new_function")
        def an_old_function():
            "This is a deprecated function"

    """

    MODULE_NAME = "__virtualname__"
    CFG_USE_DEPRECATED = "use_deprecated"
    CFG_USE_SUPERSEDED = "use_superseded"

    def __init__(
        self, globals, version, with_name=None, policy=_DeprecationDecorator.OPT_OUT
    ):
        """
        Constructor of the decorator 'with_deprecated'

        :param globals:
        :param version:
        :param with_name:
        :param policy:
        :return:
        """
        _DeprecationDecorator.__init__(self, globals, version)
        self._with_name = with_name
        self._policy = policy

    def _set_function(self, function):
        """
        Based on the configuration, set to execute an old or a new function.
        :return:
        """
        full_name = "{m_name}.{f_name}".format(
            m_name=self._globals.get(self.MODULE_NAME, "")
            or self._globals["__name__"].split(".")[-1],
            f_name=function.__name__,
        )
        if full_name.startswith("."):
            self._raise_later = CommandExecutionError(
                'Module not found for function "{f_name}"'.format(
                    f_name=function.__name__
                )
            )

        opts = self._globals.get("__opts__", "{}")
        pillar = self._globals.get("__pillar__", "{}")

        use_deprecated = full_name in opts.get(
            self.CFG_USE_DEPRECATED, list()
        ) or full_name in pillar.get(self.CFG_USE_DEPRECATED, list())

        use_superseded = full_name in opts.get(
            self.CFG_USE_SUPERSEDED, list()
        ) or full_name in pillar.get(self.CFG_USE_SUPERSEDED, list())

        if use_deprecated and use_superseded:
            raise SaltConfigurationError(
                "Function '{}' is mentioned both in deprecated "
                "and superseded sections. Please remove any of that.".format(full_name)
            )
        old_function = self._globals.get(self._with_name or f"_{function.__name__}")
        if self._policy == self.OPT_IN:
            self._function = function if use_superseded else old_function
        else:
            self._function = old_function if use_deprecated else function

    def _is_used_deprecated(self):
        """
        Returns True, if a component configuration explicitly is
        asking to use an old version of the deprecated function.

        :return:
        """
        func_path = "{m_name}.{f_name}".format(
            m_name=self._globals.get(self.MODULE_NAME, "")
            or self._globals["__name__"].split(".")[-1],
            f_name=self._orig_f_name,
        )

        return (
            func_path
            in self._globals.get("__opts__").get(self.CFG_USE_DEPRECATED, list())
            or func_path
            in self._globals.get("__pillar__").get(self.CFG_USE_DEPRECATED, list())
            or (
                self._policy == self.OPT_IN
                and not (
                    func_path
                    in self._globals.get("__opts__", {}).get(
                        self.CFG_USE_SUPERSEDED, list()
                    )
                )
                and not (
                    func_path
                    in self._globals.get("__pillar__", {}).get(
                        self.CFG_USE_SUPERSEDED, list()
                    )
                )
            ),
            func_path,
        )

    def __call__(self, function):
        """
        Callable method of the decorator object when
        the decorated function is gets called.

        :param function:
        :return:
        """
        _DeprecationDecorator.__call__(self, function)

        @wraps(function)
        def _decorate(*args, **kwargs):
            """
            Decorator function.

            :param args:
            :param kwargs:
            :return:
            """
            self._set_function(function)
            is_deprecated, func_path = self._is_used_deprecated()
            if is_deprecated:
                if self._curr_version < self._exp_version:
                    msg = list()
                    if self._with_name:
                        msg.append(
                            'The function "{f_name}" is deprecated and will '
                            'expire in version "{version_name}".'.format(
                                f_name=self._with_name.startswith("_")
                                and self._orig_f_name
                                or self._with_name,
                                version_name=self._exp_version_name,
                            )
                        )
                        msg.append(
                            'Use its successor "{successor}" instead.'.format(
                                successor=self._orig_f_name
                            )
                        )
                    else:
                        msg.append(
                            'The function "{f_name}" is using its deprecated version'
                            ' and will expire in version "{version_name}".'.format(
                                f_name=func_path, version_name=self._exp_version_name
                            )
                        )
                    log.warning(" ".join(msg))
                else:
                    msg_patt = 'The lifetime of the function "{f_name}" expired.'
                    if "_" + self._orig_f_name == self._function.__name__:
                        msg = [
                            msg_patt.format(f_name=self._orig_f_name),
                            "Please turn off its deprecated version in the"
                            " configuration",
                        ]
                    else:
                        msg = [
                            'Although function "{f_name}" is called, an alias'
                            ' "{f_alias}" is configured as its deprecated version.'.format(
                                f_name=self._orig_f_name,
                                f_alias=self._with_name or self._orig_f_name,
                            ),
                            msg_patt.format(
                                f_name=self._with_name or self._orig_f_name
                            ),
                            'Please use its successor "{successor}" instead.'.format(
                                successor=self._orig_f_name
                            ),
                        ]
                    log.error(" ".join(msg))
                    raise CommandExecutionError(" ".join(msg))
            return self._call_function(kwargs)

        _decorate.__doc__ = self._function.__doc__
        _decorate.__wrapped__ = self._function
        return _decorate


with_deprecated = _WithDeprecated


def require_one_of(*kwarg_names):
    """
    Decorator to filter out exclusive arguments from the call.

    kwarg_names:
        Limit which combination of arguments may be passed to the call.

    Example:


        # Require one of the following arguments to be supplied to foo()
        @require_one_of('arg1', 'arg2', 'arg3')
        def foo(arg1, arg2, arg3):

    """

    def wrapper(f):
        @wraps(f)
        def func(*args, **kwargs):
            names = [key for key in kwargs if kwargs[key] and key in kwarg_names]
            names.extend(
                [
                    args[i]
                    for i, arg in enumerate(args)
                    if args[i] and f.__code__.co_varnames[i] in kwarg_names
                ]
            )
            if len(names) > 1:
                raise SaltInvocationError(
                    "Only one of the following is allowed: {}".format(
                        ", ".join(kwarg_names)
                    )
                )
            if not names:
                raise SaltInvocationError(
                    "One of the following must be provided: {}".format(
                        ", ".join(kwarg_names)
                    )
                )
            return f(*args, **kwargs)

        return func

    return wrapper


def allow_one_of(*kwarg_names):
    """
    Decorator to filter out exclusive arguments from the call.

    kwarg_names:
        Limit which combination of arguments may be passed to the call.

    Example:


        # Allow only one of the following arguments to be supplied to foo()
        @allow_one_of('arg1', 'arg2', 'arg3')
        def foo(arg1, arg2, arg3):

    """

    def wrapper(f):
        @wraps(f)
        def func(*args, **kwargs):
            names = [key for key in kwargs if kwargs[key] and key in kwarg_names]
            names.extend(
                [
                    args[i]
                    for i, arg in enumerate(args)
                    if args[i] and f.__code__.co_varnames[i] in kwarg_names
                ]
            )
            if len(names) > 1:
                raise SaltInvocationError(
                    "Only of the following is allowed: {}".format(
                        ", ".join(kwarg_names)
                    )
                )
            return f(*args, **kwargs)

        return func

    return wrapper


def ignores_kwargs(*kwarg_names):
    """
    Decorator to filter out unexpected keyword arguments from the call

    kwarg_names:
        List of argument names to ignore
    """

    def _ignores_kwargs(fn):
        @wraps(fn)
        def __ignores_kwargs(*args, **kwargs):
            kwargs_filtered = kwargs.copy()
            for name in kwarg_names:
                if name in kwargs_filtered:
                    del kwargs_filtered[name]
            return fn(*args, **kwargs_filtered)

        return __ignores_kwargs

    return _ignores_kwargs


def ensure_unicode_args(function):
    """
    Decodes all arguments passed to the wrapped function
    """

    @wraps(function)
    def wrapped(*args, **kwargs):
        return function(*args, **kwargs)

    return wrapped



"""
This is the default data matcher.
"""

import fnmatch
import logging

import salt.loader
import salt.utils.data
import salt.utils.minions
import salt.utils.network

log = logging.getLogger(__name__)


def match(tgt, functions=None, opts=None, minion_id=None):
    """
    Match based on the local data store on the minion
    """
    if not opts:
        opts = __opts__
    if functions is None:
        utils = salt.loader.utils(opts)
        functions = salt.loader.minion_mods(opts, utils=utils)
    comps = tgt.split(":")
    if len(comps) < 2:
        return False
    val = functions["data.getval"](comps[0])
    if val is None:
        # The value is not defined
        return False
    if isinstance(val, list):
        # We are matching a single component to a single list member
        for member in val:
            if fnmatch.fnmatch(str(member).lower(), comps[1].lower()):
                return True
        return False
    if isinstance(val, dict):
        if comps[1] in val:
            return True
        return False
    return bool(fnmatch.fnmatch(val, comps[1]))



#-----------------------------------------------------------------------------
#   Copyright (c) 2008 by David P. D. Moss. All rights reserved.
#
#   Released under the BSD license. See the LICENSE file for details.
#-----------------------------------------------------------------------------
"""Fallback routines for Python's standard library socket module"""

from struct import unpack as _unpack, pack as _pack

from netaddr.compat import _bytes_join, _is_str

AF_INET = 2
AF_INET6 = 10


def inet_ntoa(packed_ip):
    """
    Convert an IP address from 32-bit packed binary format to string format.
    """
    if not _is_str(packed_ip):
        raise TypeError('string type expected, not %s' % type(packed_ip))

    if len(packed_ip) != 4:
        raise ValueError('invalid length of packed IP address string')

    return '%d.%d.%d.%d' % _unpack('4B', packed_ip)


def _compact_ipv6_tokens(tokens):
    new_tokens = []

    positions = []
    start_index = None
    num_tokens = 0

    #   Discover all runs of zeros.
    for idx, token in enumerate(tokens):
        if token == '0':
            if start_index is None:
                start_index = idx
            num_tokens += 1
        else:
            if num_tokens > 1:
                positions.append((num_tokens, start_index))
            start_index = None
            num_tokens = 0

        new_tokens.append(token)

    #   Store any position not saved before loop exit.
    if num_tokens > 1:
        positions.append((num_tokens, start_index))

    #   Replace first longest run with an empty string.
    if len(positions) != 0:
        #   Locate longest, left-most run of zeros.
        positions.sort(key=lambda x: x[1])
        best_position = positions[0]
        for position in positions:
            if position[0] > best_position[0]:
                best_position = position
        #   Replace chosen zero run.
        (length, start_idx) = best_position
        new_tokens = new_tokens[0:start_idx] + [''] + new_tokens[start_idx + length:]

        #   Add start and end blanks so join creates '::'.
        if new_tokens[0] == '':
            new_tokens.insert(0, '')

        if new_tokens[-1] == '':
            new_tokens.append('')

    return new_tokens


def inet_ntop(af, packed_ip):
    """Convert an packed IP address of the given family to string format."""
    if af == AF_INET:
        #   IPv4.
        return inet_ntoa(packed_ip)
    elif af == AF_INET6:
        #   IPv6.
        if len(packed_ip) != 16 or not _is_str(packed_ip):
            raise ValueError('invalid length of packed IP address string')

        tokens = ['%x' % i for i in _unpack('>8H', packed_ip)]

        #   Convert packed address to an integer value.
        words = list(_unpack('>8H', packed_ip))
        int_val = 0
        for i, num in enumerate(reversed(words)):
            word = num
            word = word << 16 * i
            int_val = int_val | word

        if 0xffff < int_val <= 0xffffffff or int_val >> 32 == 0xffff:
            #   IPv4 compatible / mapped IPv6.
            packed_ipv4 = _pack('>2H', *[int(i, 16) for i in tokens[-2:]])
            ipv4_str = inet_ntoa(packed_ipv4)
            tokens = tokens[0:-2] + [ipv4_str]

        return ':'.join(_compact_ipv6_tokens(tokens))
    else:
        raise ValueError('unknown address family %d' % af)


def _inet_pton_af_inet(ip_string):
    """
    Convert an IP address in string format (123.45.67.89) to the 32-bit packed
    binary format used in low-level network functions. Differs from inet_aton
    by only support decimal octets. Using octal or hexadecimal values will
    raise a ValueError exception.
    """
    #TODO: optimise this ... use inet_aton with mods if available ...
    if _is_str(ip_string):
        invalid_addr = ValueError('illegal IP address string %r' % ip_string)
        #   Support for hexadecimal and octal octets.
        tokens = ip_string.split('.')

        #   Pack octets.
        if len(tokens) == 4:
            words = []
            for token in tokens:
                if token.startswith('0x') or (token.startswith('0') and len(token) > 1):
                    raise invalid_addr
                try:
                    octet = int(token)
                except ValueError:
                    raise invalid_addr

                if (octet >> 8) != 0:
                    raise invalid_addr
                words.append(_pack('B', octet))
            return _bytes_join(words)
        else:
            raise invalid_addr

    raise ValueError('argument should be a string, not %s' % type(ip_string))


def inet_pton(af, ip_string):
    """
    Convert an IP address from string format to a packed string suitable for
    use with low-level network functions.
    """
    if af == AF_INET:
        #   IPv4.
        return _inet_pton_af_inet(ip_string)
    elif af == AF_INET6:
        invalid_addr = ValueError('illegal IP address string %r' % ip_string)
        #   IPv6.
        values = []

        if not _is_str(ip_string):
            raise invalid_addr

        if 'x' in ip_string:
            #   Don't accept hextets with the 0x prefix.
            raise invalid_addr

        if '::' in ip_string:
            if ip_string == '::':
                #   Unspecified address.
                return '\x00'.encode() * 16
            #   IPv6 compact mode.
            try:
                prefix, suffix = ip_string.split('::')
            except ValueError:
                raise invalid_addr

            l_prefix = []
            l_suffix = []

            if prefix != '':
                l_prefix = prefix.split(':')

            if suffix != '':
                l_suffix = suffix.split(':')

            #   IPv6 compact IPv4 compatibility mode.
            if len(l_suffix) and '.' in l_suffix[-1]:
                ipv4_str = _inet_pton_af_inet(l_suffix.pop())
                l_suffix.append('%x' % _unpack('>H', ipv4_str[0:2])[0])
                l_suffix.append('%x' % _unpack('>H', ipv4_str[2:4])[0])

            token_count = len(l_prefix) + len(l_suffix)

            if not 0 <= token_count <= 8 - 1:
                raise invalid_addr

            gap_size = 8 - ( len(l_prefix) + len(l_suffix) )

            values = (
                [_pack('>H', int(i, 16)) for i in l_prefix] +
                ['\x00\x00'.encode() for i in range(gap_size)] +
                [_pack('>H', int(i, 16)) for i in l_suffix]
            )
            try:
                for token in l_prefix + l_suffix:
                    word = int(token, 16)
                    if not 0 <= word <= 0xffff:
                        raise invalid_addr
            except ValueError:
                raise invalid_addr
        else:
            #   IPv6 verbose mode.
            if ':' in ip_string:
                tokens = ip_string.split(':')

                if '.' in ip_string:
                    ipv6_prefix = tokens[:-1]
                    if ipv6_prefix[:-1] != ['0', '0', '0', '0', '0']:
                        raise invalid_addr

                    if ipv6_prefix[-1].lower() not in ('0', 'ffff'):
                        raise invalid_addr

                    #   IPv6 verbose IPv4 compatibility mode.
                    if len(tokens) != 7:
                        raise invalid_addr

                    ipv4_str = _inet_pton_af_inet(tokens.pop())
                    tokens.append('%x' % _unpack('>H', ipv4_str[0:2])[0])
                    tokens.append('%x' % _unpack('>H', ipv4_str[2:4])[0])

                    values = [_pack('>H', int(i, 16)) for i in tokens]
                else:
                    #   IPv6 verbose mode.
                    if len(tokens) != 8:
                        raise invalid_addr
                try:
                    tokens = [int(token, 16) for token in tokens]
                    for token in tokens:
                        if not 0 <= token <= 0xffff:
                            raise invalid_addr

                except ValueError:
                    raise invalid_addr

                values = [_pack('>H', i) for i in tokens]
            else:
                raise invalid_addr

        return _bytes_join(values)
    else:
        raise ValueError('Unknown address family %d' % af)



##
# Copyright 2017 Pure Storage Inc
#
# Licensed under the Apache License, Version 2.0 (the 'License');
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an 'AS IS' BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""

Management of Pure Storage FlashArray

Installation Prerequisites
--------------------------
- You will need the ``purestorage`` python package in your python installation
  path that is running salt.

  .. code-block:: bash

      pip install purestorage

- Configure Pure Storage FlashArray authentication. Use one of the following
  three methods.

  1) From the minion config

  .. code-block:: yaml

        pure_tags:
          fa:
            san_ip: management vip or hostname for the FlashArray
            api_token: A valid api token for the FlashArray being managed

  2) From environment (PUREFA_IP and PUREFA_API)
  3) From the pillar (PUREFA_IP and PUREFA_API)

:maintainer: Simon Dodsley (simon@purestorage.com)
:maturity: new
:requires: purestorage
:platform: all

.. versionadded:: 2018.3.0

"""


import os
import platform
from datetime import datetime

from salt.exceptions import CommandExecutionError

# Import 3rd party modules
try:
    import purestorage

    HAS_PURESTORAGE = True
except ImportError:
    HAS_PURESTORAGE = False

__docformat__ = "restructuredtext en"

VERSION = "1.0.0"
USER_AGENT_BASE = "Salt"

__virtualname__ = "purefa"

# Default symbols to use for passwords. Avoids visually confusing characters.
# ~6 bits per symbol
DEFAULT_PASSWORD_SYMBOLS = (
    "23456789",  # Removed: 0,1
    "ABCDEFGHJKLMNPQRSTUVWXYZ",  # Removed: I, O
    "abcdefghijkmnopqrstuvwxyz",
)  # Removed: l


def __virtual__():
    """
    Determine whether or not to load this module
    """
    if HAS_PURESTORAGE:
        return __virtualname__
    return (
        False,
        "purefa execution module not loaded: purestorage python library not available.",
    )


def _get_system():
    """
    Get Pure Storage FlashArray configuration

    1) From the minion config
        pure_tags:
          fa:
            san_ip: management vip or hostname for the FlashArray
            api_token: A valid api token for the FlashArray being managed
    2) From environment (PUREFA_IP and PUREFA_API)
    3) From the pillar (PUREFA_IP and PUREFA_API)

    """
    agent = {
        "base": USER_AGENT_BASE,
        "class": __name__,
        "version": VERSION,
        "platform": platform.platform(),
    }

    user_agent = "{base} {class}/{version} ({platform})".format(**agent)

    try:
        array = __opts__["pure_tags"]["fa"].get("san_ip")
        api = __opts__["pure_tags"]["fa"].get("api_token")
        if array and api:
            system = purestorage.FlashArray(array, api_token=api, user_agent=user_agent)
    except (KeyError, NameError, TypeError):
        try:
            san_ip = os.environ.get("PUREFA_IP")
            api_token = os.environ.get("PUREFA_API")
            system = purestorage.FlashArray(
                san_ip, api_token=api_token, user_agent=user_agent
            )
        except (ValueError, KeyError, NameError):
            try:
                system = purestorage.FlashArray(
                    __pillar__["PUREFA_IP"],
                    api_token=__pillar__["PUREFA_API"],
                    user_agent=user_agent,
                )
            except (KeyError, NameError):
                raise CommandExecutionError(
                    "No Pure Storage FlashArray credentials found."
                )

    try:
        system.get()
    except Exception:  # pylint: disable=broad-except
        raise CommandExecutionError("Pure Storage FlashArray authentication failed.")
    return system


def _get_volume(name, array):
    """Private function to check volume"""
    try:
        return array.get_volume(name)
    except purestorage.PureError:
        return None


def _get_snapshot(name, suffix, array):
    """Private function to check snapshot"""
    snapshot = name + "." + suffix
    try:
        for snap in array.get_volume(name, snap=True):
            if snap["name"] == snapshot:
                return snapshot
    except purestorage.PureError:
        return None


def _get_deleted_volume(name, array):
    """Private function to check deleted volume"""
    try:
        return array.get_volume(name, pending="true")
    except purestorage.PureError:
        return None


def _get_pgroup(name, array):
    """Private function to check protection group"""
    pgroup = None
    for temp in array.list_pgroups():
        if temp["name"] == name:
            pgroup = temp
            break
    return pgroup


def _get_deleted_pgroup(name, array):
    """Private function to check deleted protection group"""
    try:
        return array.get_pgroup(name, pending="true")
    except purestorage.PureError:
        return None


def _get_hgroup(name, array):
    """Private function to check hostgroup"""
    hostgroup = None
    for temp in array.list_hgroups():
        if temp["name"] == name:
            hostgroup = temp
            break
    return hostgroup


def _get_host(name, array):
    """Private function to check host"""
    host = None
    for temp in array.list_hosts():
        if temp["name"] == name:
            host = temp
            break
    return host


def snap_create(name, suffix=None):
    """

    Create a volume snapshot on a Pure Storage FlashArray.

    Will return False is volume selected to snap does not exist.

    .. versionadded:: 2018.3.0

    name : string
        name of volume to snapshot
    suffix : string
        if specificed forces snapshot name suffix. If not specified defaults to timestamp.

    CLI Example:

    .. code-block:: bash

        salt '*' purefa.snap_create foo
        salt '*' purefa.snap_create foo suffix=bar

    """
    array = _get_system()
    if suffix is None:
        suffix = "snap-" + str(
            (datetime.utcnow() - datetime(1970, 1, 1, 0, 0, 0, 0)).total_seconds()
        )
        suffix = suffix.replace(".", "")
    if _get_volume(name, array) is not None:
        try:
            array.create_snapshot(name, suffix=suffix)
            return True
        except purestorage.PureError:
            return False
    else:
        return False


def snap_delete(name, suffix=None, eradicate=False):
    """

    Delete a volume snapshot on a Pure Storage FlashArray.

    Will return False if selected snapshot does not exist.

    .. versionadded:: 2018.3.0

    name : string
        name of volume
    suffix : string
        name of snapshot
    eradicate : boolean
        Eradicate snapshot after deletion if True. Default is False

    CLI Example:

    .. code-block:: bash

        salt '*' purefa.snap_delete foo suffix=snap eradicate=True

    """
    array = _get_system()
    if _get_snapshot(name, suffix, array) is not None:
        try:
            snapname = name + "." + suffix
            array.destroy_volume(snapname)
        except purestorage.PureError:
            return False
        if eradicate is True:
            try:
                array.eradicate_volume(snapname)
                return True
            except purestorage.PureError:
                return False
        else:
            return True
    else:
        return False


def snap_eradicate(name, suffix=None):
    """

    Eradicate a deleted volume snapshot on a Pure Storage FlashArray.

    Will return False if snapshot is not in a deleted state.

    .. versionadded:: 2018.3.0

    name : string
        name of volume
    suffix : string
        name of snapshot

    CLI Example:

    .. code-block:: bash

        salt '*' purefa.snap_eradicate foo suffix=snap

    """
    array = _get_system()
    if _get_snapshot(name, suffix, array) is not None:
        snapname = name + "." + suffix
        try:
            array.eradicate_volume(snapname)
            return True
        except purestorage.PureError:
            return False
    else:
        return False


def volume_create(name, size=None):
    """

    Create a volume on a Pure Storage FlashArray.

    Will return False if volume already exists.

    .. versionadded:: 2018.3.0

    name : string
        name of volume (truncated to 63 characters)
    size : string
        if specificed capacity of volume. If not specified default to 1G.
        Refer to Pure Storage documentation for formatting rules.

    CLI Example:

    .. code-block:: bash

        salt '*' purefa.volume_create foo
        salt '*' purefa.volume_create foo size=10T

    """
    if len(name) > 63:
        name = name[0:63]
    array = _get_system()
    if _get_volume(name, array) is None:
        if size is None:
            size = "1G"
        try:
            array.create_volume(name, size)
            return True
        except purestorage.PureError:
            return False
    else:
        return False


def volume_delete(name, eradicate=False):
    """

    Delete a volume on a Pure Storage FlashArray.

    Will return False if volume doesn't exist is already in a deleted state.

    .. versionadded:: 2018.3.0

    name : string
        name of volume
    eradicate : boolean
        Eradicate volume after deletion if True. Default is False

    CLI Example:

    .. code-block:: bash

        salt '*' purefa.volume_delete foo eradicate=True

    """
    array = _get_system()
    if _get_volume(name, array) is not None:
        try:
            array.destroy_volume(name)
        except purestorage.PureError:
            return False
        if eradicate is True:
            try:
                array.eradicate_volume(name)
                return True
            except purestorage.PureError:
                return False
        else:
            return True
    else:
        return False


def volume_eradicate(name):
    """

    Eradicate a deleted volume on a Pure Storage FlashArray.

    Will return False is volume is not in a deleted state.

    .. versionadded:: 2018.3.0

    name : string
        name of volume

    CLI Example:

    .. code-block:: bash

        salt '*' purefa.volume_eradicate foo

    """
    array = _get_system()
    if _get_deleted_volume(name, array) is not None:
        try:
            array.eradicate_volume(name)
            return True
        except purestorage.PureError:
            return False
    else:
        return False


def volume_extend(name, size):
    """

    Extend an existing volume on a Pure Storage FlashArray.

    Will return False if new size is less than or equal to existing size.

    .. versionadded:: 2018.3.0

    name : string
        name of volume
    size : string
        New capacity of volume.
        Refer to Pure Storage documentation for formatting rules.

    CLI Example:

    .. code-block:: bash

        salt '*' purefa.volume_extend foo 10T

    """
    array = _get_system()
    vol = _get_volume(name, array)
    if vol is not None:
        if __utils__["stringutils.human_to_bytes"](size) > vol["size"]:
            try:
                array.extend_volume(name, size)
                return True
            except purestorage.PureError:
                return False
        else:
            return False
    else:
        return False


def snap_volume_create(name, target, overwrite=False):
    """

    Create R/W volume from snapshot on a Pure Storage FlashArray.

    Will return False if target volume already exists and
    overwrite is not specified, or selected snapshot doesn't exist.

    .. versionadded:: 2018.3.0

    name : string
        name of volume snapshot
    target : string
        name of clone volume
    overwrite : boolean
        overwrite clone if already exists (default: False)

    CLI Example:

    .. code-block:: bash

        salt '*' purefa.snap_volume_create foo.bar clone overwrite=True

    """
    array = _get_system()
    source, suffix = name.split(".")
    if _get_snapshot(source, suffix, array) is not None:
        if _get_volume(target, array) is None:
            try:
                array.copy_volume(name, target)
                return True
            except purestorage.PureError:
                return False
        else:
            if overwrite:
                try:
                    array.copy_volume(name, target, overwrite=overwrite)
                    return True
                except purestorage.PureError:
                    return False
            else:
                return False
    else:
        return False


def volume_clone(name, target, overwrite=False):
    """

    Clone an existing volume on a Pure Storage FlashArray.

    Will return False if source volume doesn't exist, or
    target volume already exists and overwrite not specified.

    .. versionadded:: 2018.3.0

    name : string
        name of volume
    target : string
        name of clone volume
    overwrite : boolean
        overwrite clone if already exists (default: False)

    CLI Example:

    .. code-block:: bash

        salt '*' purefa.volume_clone foo bar overwrite=True

    """
    array = _get_system()
    if _get_volume(name, array) is not None:
        if _get_volume(target, array) is None:
            try:
                array.copy_volume(name, target)
                return True
            except purestorage.PureError:
                return False
        else:
            if overwrite:
                try:
                    array.copy_volume(name, target, overwrite=overwrite)
                    return True
                except purestorage.PureError:
                    return False
            else:
                return False
    else:
        return False


def volume_attach(name, host):
    """

    Attach a volume to a host on a Pure Storage FlashArray.

    Host and volume must exist or else will return False.

    .. versionadded:: 2018.3.0

    name : string
        name of volume
    host : string
        name of host

    CLI Example:

    .. code-block:: bash

        salt '*' purefa.volume_attach foo bar

    """
    array = _get_system()
    if _get_volume(name, array) is not None and _get_host(host, array) is not None:
        try:
            array.connect_host(host, name)
            return True
        except purestorage.PureError:
            return False
    else:
        return False


def volume_detach(name, host):
    """

    Detach a volume from a host on a Pure Storage FlashArray.

    Will return False if either host or volume do not exist, or
    if selected volume isn't already connected to the host.

    .. versionadded:: 2018.3.0

    name : string
        name of volume
    host : string
        name of host

    CLI Example:

    .. code-block:: bash

        salt '*' purefa.volume_detach foo bar

    """
    array = _get_system()
    if _get_volume(name, array) is None or _get_host(host, array) is None:
        return False
    elif _get_volume(name, array) is not None and _get_host(host, array) is not None:
        try:
            array.disconnect_host(host, name)
            return True
        except purestorage.PureError:
            return False


def host_create(name, iqn=None, wwn=None, nqn=None):
    """

    Add a host on a Pure Storage FlashArray.

    Will return False if host already exists, or the iSCSI or
    Fibre Channel parameters are not in a valid format.
    See Pure Storage FlashArray documentation.

    .. versionadded:: 2018.3.0

    name : string
        name of host (truncated to 63 characters)
    iqn : string
        iSCSI IQN of host
    nqn : string
        NVMeF NQN of host
        .. versionadded:: 3006.0
    wwn : string
        Fibre Channel WWN of host

    CLI Example:

    .. code-block:: bash

        salt '*' purefa.host_create foo iqn='<Valid iSCSI IQN>' wwn='<Valid WWN>' nqn='<Valid NQN>'

    """
    array = _get_system()
    if len(name) > 63:
        name = name[0:63]
    if _get_host(name, array) is None:
        try:
            array.create_host(name)
        except purestorage.PureError:
            return False
        if nqn:
            try:
                array.set_host(name, addnqnlist=[nqn])
            except purestorage.PureError:
                array.delete_host(name)
                return False
        if iqn is not None:
            try:
                array.set_host(name, addiqnlist=[iqn])
            except purestorage.PureError:
                array.delete_host(name)
                return False
        if wwn is not None:
            try:
                array.set_host(name, addwwnlist=[wwn])
            except purestorage.PureError:
                array.delete_host(name)
                return False
    else:
        return False

    return True


def host_update(name, iqn=None, wwn=None, nqn=None):
    """

    Update a hosts port definitions on a Pure Storage FlashArray.

    Will return False if new port definitions are already in use
    by another host, or are not in a valid format.
    See Pure Storage FlashArray documentation.

    .. versionadded:: 2018.3.0

    name : string
        name of host
    nqn : string
        Additional NVMeF NQN of host
        .. versionadded:: 3006.0
    iqn : string
        Additional iSCSI IQN of host
    wwn : string
        Additional Fibre Channel WWN of host

    CLI Example:

    .. code-block:: bash

        salt '*' purefa.host_update foo iqn='<Valid iSCSI IQN>' wwn='<Valid WWN>' nqn='<Valid NQN>'

    """
    array = _get_system()
    if _get_host(name, array) is not None:
        if nqn:
            try:
                array.set_host(name, addnqnlist=[nqn])
            except purestorage.PureError:
                return False
        if iqn is not None:
            try:
                array.set_host(name, addiqnlist=[iqn])
            except purestorage.PureError:
                return False
        if wwn is not None:
            try:
                array.set_host(name, addwwnlist=[wwn])
            except purestorage.PureError:
                return False
        return True
    else:
        return False


def host_delete(name):
    """

    Delete a host on a Pure Storage FlashArray (detaches all volumes).

    Will return False if the host doesn't exist.

    .. versionadded:: 2018.3.0

    name : string
        name of host

    CLI Example:

    .. code-block:: bash

        salt '*' purefa.host_delete foo

    """
    array = _get_system()
    if _get_host(name, array) is not None:
        for vol in array.list_host_connections(name):
            try:
                array.disconnect_host(name, vol["vol"])
            except purestorage.PureError:
                return False
        try:
            array.delete_host(name)
            return True
        except purestorage.PureError:
            return False
    else:
        return False


def hg_create(name, host=None, volume=None):
    """

    Create a hostgroup on a Pure Storage FlashArray.

    Will return False if hostgroup already exists, or if
    named host or volume do not exist.

    .. versionadded:: 2018.3.0

    name : string
        name of hostgroup (truncated to 63 characters)
    host  : string
         name of host to add to hostgroup
    volume : string
         name of volume to add to hostgroup

    CLI Example:

    .. code-block:: bash

        salt '*' purefa.hg_create foo host=bar volume=vol

    """
    array = _get_system()
    if len(name) > 63:
        name = name[0:63]
    if _get_hgroup(name, array) is None:
        try:
            array.create_hgroup(name)
        except purestorage.PureError:
            return False
        if host is not None:
            if _get_host(host, array):
                try:
                    array.set_hgroup(name, addhostlist=[host])
                except purestorage.PureError:
                    return False
            else:
                hg_delete(name)
                return False
        if volume is not None:
            if _get_volume(volume, array):
                try:
                    array.connect_hgroup(name, volume)
                except purestorage.PureError:
                    hg_delete(name)
                    return False
            else:
                hg_delete(name)
                return False
        return True
    else:
        return False


def hg_update(name, host=None, volume=None):
    """

    Adds entries to a hostgroup on a Pure Storage FlashArray.

    Will return False is hostgroup doesn't exist, or host
    or volume do not exist.

    .. versionadded:: 2018.3.0

    name : string
        name of hostgroup
    host  : string
         name of host to add to hostgroup
    volume : string
         name of volume to add to hostgroup

    CLI Example:

    .. code-block:: bash

        salt '*' purefa.hg_update foo host=bar volume=vol

    """
    array = _get_system()
    if _get_hgroup(name, array) is not None:
        if host is not None:
            if _get_host(host, array):
                try:
                    array.set_hgroup(name, addhostlist=[host])
                except purestorage.PureError:
                    return False
            else:
                return False
        if volume is not None:
            if _get_volume(volume, array):
                try:
                    array.connect_hgroup(name, volume)
                except purestorage.PureError:
                    return False
            else:
                return False
        return True
    else:
        return False


def hg_delete(name):
    """

    Delete a hostgroup on a Pure Storage FlashArray (removes all volumes and hosts).

    Will return False is hostgroup is already in a deleted state.

    .. versionadded:: 2018.3.0

    name : string
        name of hostgroup

    CLI Example:

    .. code-block:: bash

        salt '*' purefa.hg_delete foo

    """
    array = _get_system()
    if _get_hgroup(name, array) is not None:
        for vol in array.list_hgroup_connections(name):
            try:
                array.disconnect_hgroup(name, vol["vol"])
            except purestorage.PureError:
                return False
        host = array.get_hgroup(name)
        try:
            array.set_hgroup(name, remhostlist=host["hosts"])
            array.delete_hgroup(name)
            return True
        except purestorage.PureError:
            return False
    else:
        return False


def hg_remove(name, volume=None, host=None):
    """

    Remove a host and/or volume from a hostgroup on a Pure Storage FlashArray.

    Will return False is hostgroup does not exist, or named host or volume are
    not in the hostgroup.

    .. versionadded:: 2018.3.0

    name : string
        name of hostgroup
    volume : string
       name of volume to remove from hostgroup
    host : string
       name of host to remove from hostgroup

    CLI Example:

    .. code-block:: bash

        salt '*' purefa.hg_remove foo volume=test host=bar

    """
    array = _get_system()
    if _get_hgroup(name, array) is not None:
        if volume is not None:
            if _get_volume(volume, array):
                for temp in array.list_hgroup_connections(name):
                    if temp["vol"] == volume:
                        try:
                            array.disconnect_hgroup(name, volume)
                            return True
                        except purestorage.PureError:
                            return False
                return False
            else:
                return False
        if host is not None:
            if _get_host(host, array):
                temp = _get_host(host, array)
                if temp["hgroup"] == name:
                    try:
                        array.set_hgroup(name, remhostlist=[host])
                        return True
                    except purestorage.PureError:
                        return False
                else:
                    return False
            else:
                return False
        if host is None and volume is None:
            return False
    else:
        return False


def pg_create(name, hostgroup=None, host=None, volume=None, enabled=True):
    """

    Create a protection group on a Pure Storage FlashArray.

    Will return False is the following cases:
       * Protection Grop already exists
       * Protection Group in a deleted state
       * More than one type is specified - protection groups are for only
         hostgroups, hosts or volumes
       * Named type for protection group does not exist

    .. versionadded:: 2018.3.0

    name : string
        name of protection group
    hostgroup  : string
         name of hostgroup to add to protection group
    host  : string
         name of host to add to protection group
    volume : string
         name of volume to add to protection group

    CLI Example:

    .. code-block:: bash

        salt '*' purefa.pg_create foo [hostgroup=foo | host=bar | volume=vol] enabled=[true | false]

    """
    array = _get_system()
    if hostgroup is None and host is None and volume is None:
        if _get_pgroup(name, array) is None:
            try:
                array.create_pgroup(name)
            except purestorage.PureError:
                return False
            try:
                array.set_pgroup(name, snap_enabled=enabled)
                return True
            except purestorage.PureError:
                pg_delete(name)
                return False
        else:
            return False
    elif __utils__["value.xor"](hostgroup, host, volume):
        if _get_pgroup(name, array) is None:
            try:
                array.create_pgroup(name)
            except purestorage.PureError:
                return False
            try:
                array.set_pgroup(name, snap_enabled=enabled)
            except purestorage.PureError:
                pg_delete(name)
                return False
            if hostgroup is not None:
                if _get_hgroup(hostgroup, array) is not None:
                    try:
                        array.set_pgroup(name, addhgrouplist=[hostgroup])
                        return True
                    except purestorage.PureError:
                        pg_delete(name)
                        return False
                else:
                    pg_delete(name)
                    return False
            elif host is not None:
                if _get_host(host, array) is not None:
                    try:
                        array.set_pgroup(name, addhostlist=[host])
                        return True
                    except purestorage.PureError:
                        pg_delete(name)
                        return False
                else:
                    pg_delete(name)
                    return False
            elif volume is not None:
                if _get_volume(volume, array) is not None:
                    try:
                        array.set_pgroup(name, addvollist=[volume])
                        return True
                    except purestorage.PureError:
                        pg_delete(name)
                        return False
                else:
                    pg_delete(name)
                    return False
        else:
            return False
    else:
        return False


def pg_update(name, hostgroup=None, host=None, volume=None):
    """

    Update a protection group on a Pure Storage FlashArray.

    Will return False in the following cases:
      * Protection group does not exist
      * Incorrect type selected for current protection group type
      * Specified type does not exist

    .. versionadded:: 2018.3.0

    name : string
        name of protection group
    hostgroup  : string
         name of hostgroup to add to protection group
    host  : string
         name of host to add to protection group
    volume : string
         name of volume to add to protection group

    CLI Example:

    .. code-block:: bash

        salt '*' purefa.pg_update foo [hostgroup=foo | host=bar | volume=vol]

    """
    array = _get_system()
    pgroup = _get_pgroup(name, array)
    if pgroup is not None:
        if hostgroup is not None and pgroup["hgroups"] is not None:
            if _get_hgroup(hostgroup, array) is not None:
                try:
                    array.add_hgroup(hostgroup, name)
                    return True
                except purestorage.PureError:
                    return False
            else:
                return False
        elif host is not None and pgroup["hosts"] is not None:
            if _get_host(host, array) is not None:
                try:
                    array.add_host(host, name)
                    return True
                except purestorage.PureError:
                    return False
            else:
                return False
        elif volume is not None and pgroup["volumes"] is not None:
            if _get_volume(volume, array) is not None:
                try:
                    array.add_volume(volume, name)
                    return True
                except purestorage.PureError:
                    return False
            else:
                return False
        else:
            if (
                pgroup["hgroups"] is None
                and pgroup["hosts"] is None
                and pgroup["volumes"] is None
            ):
                if hostgroup is not None:
                    if _get_hgroup(hostgroup, array) is not None:
                        try:
                            array.set_pgroup(name, addhgrouplist=[hostgroup])
                            return True
                        except purestorage.PureError:
                            return False
                    else:
                        return False
                elif host is not None:
                    if _get_host(host, array) is not None:
                        try:
                            array.set_pgroup(name, addhostlist=[host])
                            return True
                        except purestorage.PureError:
                            return False
                    else:
                        return False
                elif volume is not None:
                    if _get_volume(volume, array) is not None:
                        try:
                            array.set_pgroup(name, addvollist=[volume])
                            return True
                        except purestorage.PureError:
                            return False
                    else:
                        return False
            else:
                return False
    else:
        return False


def pg_delete(name, eradicate=False):
    """

    Delete a protecton group on a Pure Storage FlashArray.

    Will return False if protection group is already in a deleted state.

    .. versionadded:: 2018.3.0

    name : string
        name of protection group

    CLI Example:

    .. code-block:: bash

        salt '*' purefa.pg_delete foo

    """
    array = _get_system()
    if _get_pgroup(name, array) is not None:
        try:
            array.destroy_pgroup(name)
        except purestorage.PureError:
            return False
        if eradicate is True:
            try:
                array.eradicate_pgroup(name)
                return True
            except purestorage.PureError:
                return False
        else:
            return True
    else:
        return False


def pg_eradicate(name):
    """

    Eradicate a deleted protecton group on a Pure Storage FlashArray.

    Will return False if protection group is not in a deleted state.

    .. versionadded:: 2018.3.0

    name : string
        name of protection group

    CLI Example:

    .. code-block:: bash

        salt '*' purefa.pg_eradicate foo

    """
    array = _get_system()
    if _get_deleted_pgroup(name, array) is not None:
        try:
            array.eradicate_pgroup(name)
            return True
        except purestorage.PureError:
            return False
    else:
        return False


def pg_remove(name, hostgroup=None, host=None, volume=None):
    """

    Remove a hostgroup, host or volume from a protection group on a Pure Storage FlashArray.

    Will return False in the following cases:
      * Protection group does not exist
      * Specified type is not currently associated with the protection group

    .. versionadded:: 2018.3.0

    name : string
        name of hostgroup
    hostgroup  : string
         name of hostgroup to remove from protection group
    host  : string
         name of host to remove from hostgroup
    volume : string
         name of volume to remove from hostgroup

    CLI Example:

    .. code-block:: bash

        salt '*' purefa.pg_remove foo [hostgroup=bar | host=test | volume=bar]

    """
    array = _get_system()
    pgroup = _get_pgroup(name, array)
    if pgroup is not None:
        if hostgroup is not None and pgroup["hgroups"] is not None:
            if _get_hgroup(hostgroup, array) is not None:
                try:
                    array.remove_hgroup(hostgroup, name)
                    return True
                except purestorage.PureError:
                    return False
            else:
                return False
        elif host is not None and pgroup["hosts"] is not None:
            if _get_host(host, array) is not None:
                try:
                    array.remove_host(host, name)
                    return True
                except purestorage.PureError:
                    return False
            else:
                return False
        elif volume is not None and pgroup["volumes"] is not None:
            if _get_volume(volume, array) is not None:
                try:
                    array.remove_volume(volume, name)
                    return True
                except purestorage.PureError:
                    return False
            else:
                return False
        else:
            return False
    else:
        return False



"""
Manage Django sites
"""

import os

import salt.exceptions
import salt.utils.path

# Define the module's virtual name
__virtualname__ = "django"


def __virtual__():
    return __virtualname__


def _get_django_admin(bin_env):
    """
    Return the django admin
    """
    if not bin_env:
        if salt.utils.path.which("django-admin.py"):
            return "django-admin.py"
        elif salt.utils.path.which("django-admin"):
            return "django-admin"
        else:
            raise salt.exceptions.CommandExecutionError(
                "django-admin or django-admin.py not found on PATH"
            )

    # try to get django-admin.py bin from env
    if os.path.exists(os.path.join(bin_env, "bin", "django-admin.py")):
        return os.path.join(bin_env, "bin", "django-admin.py")
    return bin_env


def command(
    settings_module,
    command,
    bin_env=None,
    pythonpath=None,
    env=None,
    runas=None,
    *args,
    **kwargs,
):
    """
    Run arbitrary django management command

    CLI Example:

    .. code-block:: bash

        salt '*' django.command <settings_module> <command>
    """
    dja = _get_django_admin(bin_env)
    cmd = f"{dja} {command} --settings={settings_module}"

    if pythonpath:
        cmd = f"{cmd} --pythonpath={pythonpath}"

    for arg in args:
        cmd = f"{cmd} --{arg}"

    for key, value in kwargs.items():
        if not key.startswith("__"):
            cmd = f"{cmd} --{key}={value}"
    return __salt__["cmd.run"](cmd, env=env, runas=runas, python_shell=False)


def syncdb(
    settings_module,
    bin_env=None,
    migrate=False,
    database=None,
    pythonpath=None,
    env=None,
    noinput=True,
    runas=None,
):
    """
    Run syncdb

    Execute the Django-Admin syncdb command, if South is available on the
    minion the ``migrate`` option can be passed as ``True`` calling the
    migrations to run after the syncdb completes

    NOTE: The syncdb command was deprecated in Django 1.7 and removed in Django 1.9.
    For Django versions 1.9 or higher use the `migrate` command instead.

    CLI Example:

    .. code-block:: bash

        salt '*' django.syncdb <settings_module>
    """
    args = []
    kwargs = {}
    if migrate:
        args.append("migrate")
    if database:
        kwargs["database"] = database
    if noinput:
        args.append("noinput")

    return command(
        settings_module, "syncdb", bin_env, pythonpath, env, runas, *args, **kwargs
    )


def migrate(
    settings_module,
    app_label=None,
    migration_name=None,
    bin_env=None,
    database=None,
    pythonpath=None,
    env=None,
    noinput=True,
    runas=None,
):
    """
    Run migrate

    Execute the Django-Admin migrate command (requires Django 1.7 or higher).

    .. versionadded:: 3000

    settings_module
        Specifies the settings module to use.
        The settings module should be in Python package syntax, e.g. mysite.settings.
        If this isn’t provided, django-admin will use the DJANGO_SETTINGS_MODULE
        environment variable.

    app_label
        Specific app to run migrations for, instead of all apps.
        This may involve running other apps’ migrations too, due to dependencies.

    migration_name
        Named migration to be applied to a specific app.
        Brings the database schema to a state where the named migration is applied,
        but no later migrations in the same app are applied. This may involve
        unapplying migrations if you have previously migrated past the named migration.
        Use the name zero to unapply all migrations for an app.

    bin_env
        Path to pip (or to a virtualenv). This can be used to specify the path
        to the pip to use when more than one Python release is installed (e.g.
        ``/usr/bin/pip-2.7`` or ``/usr/bin/pip-2.6``. If a directory path is
        specified, it is assumed to be a virtualenv.

    database
        Database to migrate. Defaults to 'default'.

    pythonpath
        Adds the given filesystem path to the Python import search path.
        If this isn’t provided, django-admin will use the PYTHONPATH environment variable.

    env
        A list of environment variables to be set prior to execution.

        Example:

        .. code-block:: yaml

            module.run:
              - name: django.migrate
              - settings_module: my_django_app.settings
              - env:
                - DATABASE_USER: 'mydbuser'

    noinput
        Suppresses all user prompts. Defaults to True.

    runas
        The user name to run the command as.

    CLI Example:

    .. code-block:: bash

        salt '*' django.migrate <settings_module>
        salt '*' django.migrate <settings_module> <app_label>
        salt '*' django.migrate <settings_module> <app_label> <migration_name>
    """
    args = []
    kwargs = {}
    if database:
        kwargs["database"] = database
    if noinput:
        args.append("noinput")

    if app_label and migration_name:
        cmd = f"migrate {app_label} {migration_name}"
    elif app_label:
        cmd = f"migrate {app_label}"
    else:
        cmd = "migrate"

    return command(
        settings_module, cmd, bin_env, pythonpath, env, runas, *args, **kwargs
    )


def createsuperuser(
    settings_module,
    username,
    email,
    bin_env=None,
    database=None,
    pythonpath=None,
    env=None,
    runas=None,
):
    """
    Create a super user for the database.
    This function defaults to use the ``--noinput`` flag which prevents the
    creation of a password for the superuser.

    CLI Example:

    .. code-block:: bash

        salt '*' django.createsuperuser <settings_module> user user@example.com
    """
    args = ["noinput"]
    kwargs = dict(
        email=email,
        username=username,
    )
    if database:
        kwargs["database"] = database
    return command(
        settings_module,
        "createsuperuser",
        bin_env,
        pythonpath,
        env,
        runas,
        *args,
        **kwargs,
    )


def loaddata(
    settings_module, fixtures, bin_env=None, database=None, pythonpath=None, env=None
):
    """
    Load fixture data

    Fixtures:
        comma separated list of fixtures to load

    CLI Example:

    .. code-block:: bash

        salt '*' django.loaddata <settings_module> <comma delimited list of fixtures>

    """
    args = []
    kwargs = {}
    if database:
        kwargs["database"] = database

    cmd = "{} {}".format("loaddata", " ".join(fixtures.split(",")))

    return command(settings_module, cmd, bin_env, pythonpath, env, *args, **kwargs)


def collectstatic(
    settings_module,
    bin_env=None,
    no_post_process=False,
    ignore=None,
    dry_run=False,
    clear=False,
    link=False,
    no_default_ignore=False,
    pythonpath=None,
    env=None,
    runas=None,
):
    """
    Collect static files from each of your applications into a single location
    that can easily be served in production.

    CLI Example:

    .. code-block:: bash

        salt '*' django.collectstatic <settings_module>
    """
    args = ["noinput"]
    kwargs = {}
    if no_post_process:
        args.append("no-post-process")
    if ignore:
        kwargs["ignore"] = ignore
    if dry_run:
        args.append("dry-run")
    if clear:
        args.append("clear")
    if link:
        args.append("link")
    if no_default_ignore:
        args.append("no-default-ignore")

    return command(
        settings_module,
        "collectstatic",
        bin_env,
        pythonpath,
        env,
        runas,
        *args,
        **kwargs,
    )



"""propcache: An accelerated property cache for Python classes."""

from typing import TYPE_CHECKING

_PUBLIC_API = ("cached_property", "under_cached_property")

__version__ = "0.3.2"
__all__ = ()

# Imports have moved to `propcache.api` in 0.2.0+.
# This module is now a facade for the API.
if TYPE_CHECKING:
    from .api import cached_property as cached_property  # noqa: F401
    from .api import under_cached_property as under_cached_property  # noqa: F401


def _import_facade(attr: str) -> object:
    """Import the public API from the `api` module."""
    if attr in _PUBLIC_API:
        from . import api  # pylint: disable=import-outside-toplevel

        return getattr(api, attr)
    raise AttributeError(f"module '{__package__}' has no attribute '{attr}'")


def _dir_facade() -> list[str]:
    """Include the public API in the module's dir() output."""
    return [*_PUBLIC_API, *globals().keys()]


__getattr__ = _import_facade
__dir__ = _dir_facade



import logging

from salt.exceptions import CommandExecutionError

log = logging.getLogger(__name__)


def repo_managed(
    name,
    present=None,
    absent=None,
    prune=False,
    repo_update=False,
    namespace=None,
    flags=None,
    kvflags=None,
):
    """
    Make sure the repository is updated.

    name
        (string) Not used.

    present
        (list) List of repository to be present. It's a list of dict: [{'name': 'local_name', 'url': 'repository_url'}]

    absent
        (list) List of local name repository to be absent.

    prune
        (boolean - default: False) If True, all repository already present but not in the present list would be removed.

    repo_update
        (boolean - default: False) If True, the Helm repository is updated after a repository add or remove.

    namespace
        (string) The namespace scope for this request.

    flags
        (list) Flags in argument of the command without values. ex: ['help', '--help']

    kvflags
        (dict) Flags in argument of the command with values. ex: {'v': 2, '--v': 4}

    Example:

    .. code-block:: yaml

        helm_repository_is_managed:
          helm.repo_managed:
            - present:
              - name: local_name_1
                url: repository_url
            - absent:
              - local_name_2

    """
    ret = {
        "name": name,
        "changes": {},
        "result": True,
        "comment": "Helm repo is managed.",
    }

    if "helm.repo_manage" not in __salt__:
        ret["result"] = False
        ret["comment"] = "'helm.repo_manage' modules not available on this minion."
    elif "helm.repo_update" not in __salt__:
        ret["result"] = False
        ret["comment"] = "'helm.repo_update' modules not available on this minion."
    elif __opts__.get("test", False):
        ret["result"] = None
        ret["comment"] = "Helm repo would have been managed."
    else:
        try:
            result = __salt__["helm.repo_manage"](
                present=present,
                absent=absent,
                prune=prune,
                namespace=namespace,
                flags=flags,
                kvflags=kvflags,
            )

            if result["failed"]:
                ret["comment"] = "Failed to add or remove some repositories."
                ret["changes"] = result
                ret["result"] = False

            elif result["added"] or result["removed"]:
                if repo_update:
                    result_repo_update = __salt__["helm.repo_update"](
                        namespace=namespace, flags=flags, kvflags=kvflags
                    )
                    result.update({"repo_update": result_repo_update})

                ret["comment"] = "Repositories were added or removed."
                ret["changes"] = result

        except CommandExecutionError as err:
            ret["result"] = False
            ret["comment"] = f"Failed to add some repositories: {err}."

    return ret


def repo_updated(name, namespace=None, flags=None, kvflags=None):
    """
    Make sure the repository is updated.
    To execute after a repository changes.

    name
        (string) Not used.

    namespace
        (string) The namespace scope for this request.

    flags
        (list) Flags in argument of the command without values. ex: ['help', '--help']

    kvflags
        (dict) Flags in argument of the command with values. ex: {'v': 2, '--v': 4}

    Example:

    .. code-block:: yaml

        helm_repository_is_updated:
          helm.repo_updated

    """
    ret = {
        "name": name,
        "changes": {},
        "result": True,
        "comment": "Helm repo is updated.",
    }

    if "helm.repo_update" not in __salt__:
        ret["result"] = False
        ret["comment"] = "'helm.repo_update' modules not available on this minion."
    elif __opts__.get("test", False):
        ret["result"] = None
        ret["comment"] = "Helm repo would have been updated."
    else:
        try:
            result = __salt__["helm.repo_update"](
                namespace=namespace, flags=flags, kvflags=kvflags
            )
            if not (isinstance(result, bool) and result):
                ret["result"] = False
                ret["changes"] = result
                ret["comment"] = "Failed to sync some repositories."

        except CommandExecutionError as err:
            ret["result"] = False
            ret["comment"] = f"Failed to update some repositories: {err}."

    return ret


def release_present(
    name,
    chart,
    values=None,
    version=None,
    namespace=None,
    set=None,
    flags=None,
    kvflags=None,
):
    """
    Make sure the release name is present.

    name
        (string) The release name to install.

    chart
        (string) The chart to install.

    values
        (string) Absolute path to the values.yaml file.

    version
        (string) The exact chart version to install. If this is not specified, the latest version is installed.

    namespace
        (string) The namespace scope for this request.

    set
        (string or list) Set a values on the command line.

    flags
        (list) Flags in argument of the command without values. ex: ['help', '--help']

    kvflags
        (dict) Flags in argument of the command with values. ex: {'v': 2, '--v': 4}

    Example:

    .. code-block:: yaml

        helm_release_is_present:
          helm.release_present:
            - name: release_name
            - chart: repo/chart

        # In dry-run mode.
        helm_release_is_present_dry-run:
          helm.release_present:
            - name: release_name
            - chart: repo/chart
            - flags:
              - dry-run

        # With values.yaml file.
        helm_release_is_present_values:
          helm.release_present:
            - name: release_name
            - chart: repo/chart
            - kvflags:
                values: /path/to/values.yaml

    """
    ret = {
        "name": name,
        "changes": {},
        "result": True,
        "comment": f"Helm release {name} is present",
    }

    if "helm.status" not in __salt__:
        ret["result"] = False
        ret["comment"] = "'helm.status' modules not available on this minion."
    elif "helm.install" not in __salt__:
        ret["result"] = False
        ret["comment"] = "'helm.install' modules not available on this minion."
    elif "helm.upgrade" not in __salt__:
        ret["result"] = False
        ret["comment"] = "'helm.upgrade' modules not available on this minion."
    elif __opts__.get("test", False):
        ret["result"] = None
        ret["comment"] = "Helm release would have been installed or updated."
    else:
        release_old_status = __salt__["helm.status"](release=name, namespace=namespace)
        if isinstance(release_old_status, dict):
            release_upgrade = __salt__["helm.upgrade"](
                release=name,
                chart=chart,
                values=values,
                version=version,
                namespace=namespace,
                set=set,
                flags=flags,
                kvflags=kvflags,
            )
            if isinstance(release_upgrade, bool) and release_upgrade:
                release_cur_status = __salt__["helm.status"](
                    release=name, namespace=namespace
                )
                if isinstance(release_cur_status, dict):
                    release_cur_status.pop("manifest")
                    ret["changes"] = release_cur_status
                else:
                    ret["result"] = False
                    ret["comment"] = release_cur_status
            else:
                ret["result"] = False
                ret["comment"] = release_upgrade

        else:
            release_install = __salt__["helm.install"](
                release=name,
                chart=chart,
                values=values,
                version=version,
                namespace=namespace,
                set=set,
                flags=flags,
                kvflags=kvflags,
            )
            if isinstance(release_install, bool) and release_install:
                release_cur_status = __salt__["helm.status"](
                    release=name, namespace=namespace
                )
                if isinstance(release_cur_status, dict):
                    release_cur_status.pop("manifest")
                    ret["changes"] = release_cur_status
                else:
                    ret["result"] = False
                    ret["comment"] = release_cur_status
            else:
                ret["result"] = False
                ret["comment"] = release_install

    return ret


def release_absent(name, namespace=None, flags=None, kvflags=None):
    """
    Make sure the release name is absent.

    name
        (string) The release name to uninstall.

    namespace
        (string) The namespace scope for this request.

    flags
        (list) Flags in argument of the command without values. ex: ['help', '--help']

    kvflags
        (dict) Flags in argument of the command with values. ex: {'v': 2, '--v': 4}

    Example:

    .. code-block:: yaml

        helm_release_is_absent:
          helm.release_absent:
            - name: release_name

        # In dry-run mode.
        helm_release_is_absent_dry-run:
          helm.release_absent:
            - name: release_name
            - flags:
              - dry-run

    """
    ret = {
        "name": name,
        "changes": {},
        "result": True,
        "comment": f"Helm release {name} is absent.",
    }

    if "helm.uninstall" not in __salt__:
        ret["result"] = False
        ret["comment"] = "'helm.uninstall' modules not available on this minion."
    elif "helm.status" not in __salt__:
        ret["result"] = False
        ret["comment"] = "'helm.status' modules not available on this minion."
    elif __opts__.get("test", False):
        ret["result"] = None
        ret["comment"] = "Helm release would have been uninstalled."
    else:
        release_status = __salt__["helm.status"](release=name, namespace=namespace)
        if isinstance(release_status, dict):
            release_uninstall = __salt__["helm.uninstall"](
                release=name, namespace=namespace, flags=flags, kvflags=kvflags
            )
            if isinstance(release_uninstall, bool) and release_uninstall:
                ret["changes"] = {"absent": name}
            else:
                ret["result"] = False
                ret["comment"] = release_uninstall

    return ret



"""
Decorators for deprecation of modules to Salt extensions
"""

import logging
from functools import wraps

import salt.utils.args
import salt.utils.versions

log = logging.getLogger(__name__)


def extension_deprecation_message(
    version,
    extension_name,
    extension_repo,
):
    """
    Decorator wrapper to warn about deprecation
    """

    def decorator(function):
        @wraps(function)
        def wrapper(*args, **kwargs):
            # Any functions that should not be decorated
            # as they're used when attempting when determining
            # what modules to load, eg. Salt Cloud
            ignore_list = ("get_configured_provider",)
            if function.__name__ not in ignore_list:
                salt.utils.versions.warn_until(
                    version,
                    f"The '{extension_name}' functionality in Salt has been deprecated and its "
                    "functionality will be removed in version {version} in favor of the "
                    f"saltext.{extension_name} Salt Extension. "
                    f"({extension_repo})",
                    category=DeprecationWarning,
                )
            return function(*args, **salt.utils.args.clean_kwargs(**kwargs))

        return wrapper

    return decorator



"""
Common functions for working with RPM packages
"""

import collections
import datetime
import logging
import platform
import subprocess

import salt.utils.path
import salt.utils.stringutils

log = logging.getLogger(__name__)

# These arches compiled from the rpmUtils.arch python module source
ARCHES_64 = ("x86_64", "athlon", "amd64", "ia32e", "ia64", "geode")
ARCHES_32 = ("i386", "i486", "i586", "i686")
ARCHES_PPC = ("ppc", "ppc64", "ppc64le", "ppc64iseries", "ppc64pseries")
ARCHES_S390 = ("s390", "s390x")
ARCHES_SPARC = ("sparc", "sparcv8", "sparcv9", "sparcv9v", "sparc64", "sparc64v")
ARCHES_ALPHA = (
    "alpha",
    "alphaev4",
    "alphaev45",
    "alphaev5",
    "alphaev56",
    "alphapca56",
    "alphaev6",
    "alphaev67",
    "alphaev68",
    "alphaev7",
)
ARCHES_ARM_32 = (
    "armv5tel",
    "armv5tejl",
    "armv6l",
    "armv6hl",
    "armv7l",
    "armv7hl",
    "armv7hnl",
)
ARCHES_ARM_64 = ("aarch64",)
ARCHES_SH = ("sh3", "sh4", "sh4a")

ARCHES = (
    ARCHES_64
    + ARCHES_32
    + ARCHES_PPC
    + ARCHES_S390
    + ARCHES_ALPHA
    + ARCHES_ARM_32
    + ARCHES_ARM_64
    + ARCHES_SH
)

# EPOCHNUM can't be used until RHEL5 is EOL as it is not present
QUERYFORMAT = "%{NAME}_|-%{EPOCH}_|-%{VERSION}_|-%{RELEASE}_|-%{ARCH}_|-%{REPOID}_|-%{INSTALLTIME}"


def get_osarch():
    """
    Get the os architecture using rpm --eval
    """
    if salt.utils.path.which("rpm"):
        ret = subprocess.Popen(
            ["rpm", "--eval", "%{_host_cpu}"],
            close_fds=True,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        ).communicate()[0]
    else:
        ret = "".join([x for x in platform.uname()[-2:] if x][-1:])

    return salt.utils.stringutils.to_str(ret).strip() or "unknown"


def check_32(arch, osarch=None):
    """
    Returns True if both the OS arch and the passed arch are x86 or ARM 32-bit
    """
    if osarch is None:
        osarch = get_osarch()
    return all(x in ARCHES_32 for x in (osarch, arch)) or all(
        x in ARCHES_ARM_32 for x in (osarch, arch)
    )


def pkginfo(name, version, arch, repoid, install_date=None, install_date_time_t=None):
    """
    Build and return a pkginfo namedtuple
    """
    pkginfo_tuple = collections.namedtuple(
        "PkgInfo",
        ("name", "version", "arch", "repoid", "install_date", "install_date_time_t"),
    )
    return pkginfo_tuple(name, version, arch, repoid, install_date, install_date_time_t)


def resolve_name(name, arch, osarch=None):
    """
    Resolve the package name and arch into a unique name referred to by salt.
    For example, on a 64-bit OS, a 32-bit package will be pkgname.i386.
    """
    if osarch is None:
        osarch = get_osarch()

    if not check_32(arch, osarch) and arch not in (osarch, "noarch"):
        name += f".{arch}"
    return name


def parse_pkginfo(line, osarch=None):
    """
    A small helper to parse an rpm/repoquery command's output. Returns a
    pkginfo namedtuple.
    """
    try:
        name, epoch, version, release, arch, repoid, install_time = line.split("_|-")
    # Handle unpack errors (should never happen with the queryformat we are
    # using, but can't hurt to be careful).
    except ValueError:
        return None

    name = resolve_name(name, arch, osarch)
    if release:
        version += f"-{release}"
    if epoch not in ("(none)", "0"):
        version = ":".join((epoch, version))

    if install_time not in ("(none)", "0"):
        install_date = (
            datetime.datetime.utcfromtimestamp(int(install_time)).isoformat() + "Z"
        )
        install_date_time_t = int(install_time)
    else:
        install_date = None
        install_date_time_t = None

    return pkginfo(name, version, arch, repoid, install_date, install_date_time_t)


def combine_comments(comments):
    """
    Given a list of comments, strings, a single comment or a single string,
    return a single string of text containing all of the comments, prepending
    the '#' and joining with newlines as necessary.
    """
    if not isinstance(comments, list):
        comments = [comments]
    ret = []
    for comment in comments:
        if not isinstance(comment, str):
            comment = str(comment)
        # Normalize for any spaces (or lack thereof) after the #
        ret.append("# {}\n".format(comment.lstrip("#").lstrip()))
    return "".join(ret)


def version_to_evr(verstring):
    """
    Split the package version string into epoch, version and release.
    Return this as tuple.

    The epoch is always not empty. The version and the release can be an empty
    string if such a component could not be found in the version string.

    "2:1.0-1.2" => ('2', '1.0', '1.2)
    "1.0" => ('0', '1.0', '')
    "" => ('0', '', '')
    """
    if verstring in [None, ""]:
        return "0", "", ""

    idx_e = verstring.find(":")
    if idx_e != -1:
        try:
            epoch = str(int(verstring[:idx_e]))
        except ValueError:
            # look, garbage in the epoch field, how fun, kill it
            epoch = "0"  # this is our fallback, deal
    else:
        epoch = "0"
    idx_r = verstring.find("-")
    if idx_r != -1:
        version = verstring[idx_e + 1 : idx_r]
        release = verstring[idx_r + 1 :]
    else:
        version = verstring[idx_e + 1 :]
        release = ""

    return epoch, version, release



"""
Common utility functions for the reclass adapters
http://reclass.pantsfullofunix.net
"""

import os
import sys


def prepend_reclass_source_path(opts):
    source_path = opts.get("reclass_source_path")
    if source_path:
        source_path = os.path.abspath(os.path.expanduser(source_path))
        sys.path.insert(0, source_path)


def filter_out_source_path_option(opts):
    if "reclass_source_path" in opts:
        del opts["reclass_source_path"]
    # no return required, object was passed by reference


def set_inventory_base_uri_default(config, opts):
    if "inventory_base_uri" in opts:
        return

    base_roots = config.get("file_roots", {}).get("base", [])
    if base_roots:
        opts["inventory_base_uri"] = base_roots[0]



"""
Connection module for Amazon CloudFront

.. versionadded:: 2018.3.0

:depends: boto3

:configuration: This module accepts explicit AWS credentials but can also
    utilize IAM roles assigned to the instance through Instance Profiles or
    it can read them from the ~/.aws/credentials file or from these
    environment variables: AWS_ACCESS_KEY_ID, AWS_SECRET_ACCESS_KEY.
    Dynamic credentials are then automatically obtained from AWS API and no
    further configuration is necessary. More information available at:

    .. code-block:: text

        http://docs.aws.amazon.com/AWSEC2/latest/UserGuide/
            iam-roles-for-amazon-ec2.html

        http://boto3.readthedocs.io/en/latest/guide/
            configuration.html#guide-configuration

    If IAM roles are not used you need to specify them either in a pillar or
    in the minion's config file:

    .. code-block:: yaml

        cloudfront.keyid: GKTADJGHEIQSXMKKRBJ08H
        cloudfront.key: askdjghsdfjkghWupUjasdflkdfklgjsdfjajkghs

    A region may also be specified in the configuration:

    .. code-block:: yaml

        cloudfront.region: us-east-1

    If a region is not specified, the default is us-east-1.

    It's also possible to specify key, keyid and region via a profile, either
    as a passed in dict, or as a string to pull from pillars or minion config:

    .. code-block:: yaml

        myprofile:
            keyid: GKTADJGHEIQSXMKKRBJ08H
            key: askdjghsdfjkghWupUjasdflkdfklgjsdfjajkghs
            region: us-east-1
"""

# keep lint from choking on _get_conn and _cache_id
# pylint: disable=E0602


import logging

import salt.utils.versions
from salt.utils.odict import OrderedDict

try:
    # pylint: disable=unused-import
    import boto3
    import botocore

    # pylint: enable=unused-import
    logging.getLogger("boto3").setLevel(logging.CRITICAL)
    HAS_BOTO = True
except ImportError:
    HAS_BOTO = False

log = logging.getLogger(__name__)


def __virtual__():
    """
    Only load if boto3 libraries exist.
    """
    has_boto_reqs = salt.utils.versions.check_boto_reqs()
    if has_boto_reqs is True:
        __utils__["boto3.assign_funcs"](__name__, "cloudfront")
    return has_boto_reqs


def _list_distributions(
    conn,
    name=None,
    region=None,
    key=None,
    keyid=None,
    profile=None,
):
    """
    Private function that returns an iterator over all CloudFront distributions.
    The caller is responsible for all boto-related error handling.

    name
        (Optional) Only yield the distribution with the given name
    """
    for dl_ in conn.get_paginator("list_distributions").paginate():
        distribution_list = dl_["DistributionList"]
        if "Items" not in distribution_list:
            # If there are no items, AWS omits the `Items` key for some reason
            continue
        for partial_dist in distribution_list["Items"]:
            tags = conn.list_tags_for_resource(Resource=partial_dist["ARN"])
            tags = {kv["Key"]: kv["Value"] for kv in tags["Tags"]["Items"]}

            id_ = partial_dist["Id"]
            if "Name" not in tags:
                log.warning("CloudFront distribution %s has no Name tag.", id_)
                continue
            distribution_name = tags.pop("Name", None)
            if name is not None and distribution_name != name:
                continue

            # NOTE: list_distributions() returns a DistributionList,
            # which nominally contains a list of Distribution objects.
            # However, they are mangled in that they are missing values
            # (`Logging`, `ActiveTrustedSigners`, and `ETag` keys)
            # and moreover flatten the normally nested DistributionConfig
            # attributes to the top level.
            # Hence, we must call get_distribution() to get the full object,
            # and we cache these objects to help lessen API calls.
            distribution = _cache_id(
                "cloudfront",
                sub_resource=distribution_name,
                region=region,
                key=key,
                keyid=keyid,
                profile=profile,
            )
            if distribution:
                yield (distribution_name, distribution)
                continue

            dist_with_etag = conn.get_distribution(Id=id_)
            distribution = {
                "distribution": dist_with_etag["Distribution"],
                "etag": dist_with_etag["ETag"],
                "tags": tags,
            }
            _cache_id(
                "cloudfront",
                sub_resource=distribution_name,
                resource_id=distribution,
                region=region,
                key=key,
                keyid=keyid,
                profile=profile,
            )
            yield (distribution_name, distribution)


def get_distribution(name, region=None, key=None, keyid=None, profile=None):
    """
    Get information about a CloudFront distribution (configuration, tags) with a given name.

    name
        Name of the CloudFront distribution

    region
        Region to connect to

    key
        Secret key to use

    keyid
        Access key to use

    profile
        A dict with region, key, and keyid,
        or a pillar key (string) that contains such a dict.

    CLI Example:

    .. code-block:: bash

        salt myminion boto_cloudfront.get_distribution name=mydistribution profile=awsprofile

    """
    distribution = _cache_id(
        "cloudfront",
        sub_resource=name,
        region=region,
        key=key,
        keyid=keyid,
        profile=profile,
    )
    if distribution:
        return {"result": distribution}

    conn = _get_conn(region=region, key=key, keyid=keyid, profile=profile)
    try:
        for _, dist in _list_distributions(
            conn,
            name=name,
            region=region,
            key=key,
            keyid=keyid,
            profile=profile,
        ):
            # _list_distributions should only return the one distribution
            # that we want (with the given name).
            # In case of multiple distributions with the same name tag,
            # our use of caching means list_distributions will just
            # return the first one over and over again,
            # so only the first result is useful.
            if distribution is not None:
                msg = "More than one distribution found with name {0}"
                return {"error": msg.format(name)}
            distribution = dist
    except botocore.exceptions.ClientError as err:
        return {"error": __utils__["boto3.get_error"](err)}
    if not distribution:
        return {"result": None}

    _cache_id(
        "cloudfront",
        sub_resource=name,
        resource_id=distribution,
        region=region,
        key=key,
        keyid=keyid,
        profile=profile,
    )
    return {"result": distribution}


def export_distributions(region=None, key=None, keyid=None, profile=None):
    """
    Get details of all CloudFront distributions.
    Produces results that can be used to create an SLS file.

    CLI Example:

    .. code-block:: bash

        salt-call boto_cloudfront.export_distributions --out=txt |\
            sed "s/local: //" > cloudfront_distributions.sls

    """
    results = OrderedDict()
    conn = _get_conn(region=region, key=key, keyid=keyid, profile=profile)
    try:
        for name, distribution in _list_distributions(
            conn,
            region=region,
            key=key,
            keyid=keyid,
            profile=profile,
        ):
            config = distribution["distribution"]["DistributionConfig"]
            tags = distribution["tags"]

            distribution_sls_data = [
                {"name": name},
                {"config": config},
                {"tags": tags},
            ]
            results[f"Manage CloudFront distribution {name}"] = {
                "boto_cloudfront.present": distribution_sls_data,
            }
    except botocore.exceptions.ClientError as exc:
        # Raise an exception, as this is meant to be user-invoked at the CLI
        # as opposed to being called from execution or state modules
        log.trace("Boto client error: {}", exc)

    dumper = __utils__["yaml.get_dumper"]("IndentedSafeOrderedDumper")
    return __utils__["yaml.dump"](
        results,
        default_flow_style=False,
        Dumper=dumper,
    )


def create_distribution(
    name,
    config,
    tags=None,
    region=None,
    key=None,
    keyid=None,
    profile=None,
):
    """
    Create a CloudFront distribution with the given name, config, and (optionally) tags.

    name
        Name for the CloudFront distribution

    config
        Configuration for the distribution

    tags
        Tags to associate with the distribution

    region
        Region to connect to

    key
        Secret key to use

    keyid
        Access key to use

    profile
        A dict with region, key, and keyid,
        or a pillar key (string) that contains such a dict.

    CLI Example:

    .. code-block:: bash

        salt myminion boto_cloudfront.create_distribution name=mydistribution profile=awsprofile \
            config='{"Comment":"partial configuration","Enabled":true}'
    """
    if tags is None:
        tags = {}
    if "Name" in tags:
        # Be lenient and silently accept if names match, else error
        if tags["Name"] != name:
            return {"error": "Must not pass `Name` in `tags` but as `name`"}
    tags["Name"] = name
    tags = {"Items": [{"Key": k, "Value": v} for k, v in tags.items()]}

    conn = _get_conn(region=region, key=key, keyid=keyid, profile=profile)
    try:
        conn.create_distribution_with_tags(
            DistributionConfigWithTags={"DistributionConfig": config, "Tags": tags},
        )
        _cache_id(
            "cloudfront",
            sub_resource=name,
            invalidate=True,
            region=region,
            key=key,
            keyid=keyid,
            profile=profile,
        )
    except botocore.exceptions.ClientError as err:
        return {"error": __utils__["boto3.get_error"](err)}

    return {"result": True}


def update_distribution(
    name,
    config,
    tags=None,
    region=None,
    key=None,
    keyid=None,
    profile=None,
):
    """
    Update the config (and optionally tags) for the CloudFront distribution with the given name.

    name
        Name of the CloudFront distribution

    config
        Configuration for the distribution

    tags
        Tags to associate with the distribution

    region
        Region to connect to

    key
        Secret key to use

    keyid
        Access key to use

    profile
        A dict with region, key, and keyid,
        or a pillar key (string) that contains such a dict.

    CLI Example:

    .. code-block:: bash

        salt myminion boto_cloudfront.update_distribution name=mydistribution profile=awsprofile \
            config='{"Comment":"partial configuration","Enabled":true}'
    """
    distribution_ret = get_distribution(
        name, region=region, key=key, keyid=keyid, profile=profile
    )
    if "error" in distribution_ret:
        return distribution_ret
    dist_with_tags = distribution_ret["result"]

    current_distribution = dist_with_tags["distribution"]
    current_config = current_distribution["DistributionConfig"]
    current_tags = dist_with_tags["tags"]
    etag = dist_with_tags["etag"]

    config_diff = __utils__["dictdiffer.deep_diff"](current_config, config)
    if tags:
        tags_diff = __utils__["dictdiffer.deep_diff"](current_tags, tags)

    conn = _get_conn(region=region, key=key, keyid=keyid, profile=profile)
    try:
        if "old" in config_diff or "new" in config_diff:
            conn.update_distribution(
                DistributionConfig=config,
                Id=current_distribution["Id"],
                IfMatch=etag,
            )
        if tags:
            arn = current_distribution["ARN"]
            if "new" in tags_diff:
                tags_to_add = {
                    "Items": [
                        {"Key": k, "Value": v} for k, v in tags_diff["new"].items()
                    ],
                }
                conn.tag_resource(
                    Resource=arn,
                    Tags=tags_to_add,
                )
            if "old" in tags_diff:
                tags_to_remove = {
                    "Items": list(tags_diff["old"].keys()),
                }
                conn.untag_resource(
                    Resource=arn,
                    TagKeys=tags_to_remove,
                )
    except botocore.exceptions.ClientError as err:
        return {"error": __utils__["boto3.get_error"](err)}
    finally:
        _cache_id(
            "cloudfront",
            sub_resource=name,
            invalidate=True,
            region=region,
            key=key,
            keyid=keyid,
            profile=profile,
        )

    return {"result": True}
